from typing import Callable
from typing import Dict
from typing import Generator
from typing import Tuple
from typing import TypeVar

import networkx as nx

from ._base import NetworkxTransformBase
from caldera.utils import functional as Fn
from caldera.utils.nx import nx_copy

T = TypeVar("T")
TupleGen = Generator[Tuple, None, None]
GraphGen = Generator[nx.Graph, None, None]


class NetworkxTransformFeatures(NetworkxTransformBase):
    def __init__(
        self,
        node_transform: Callable[[TupleGen], TupleGen] = None,
        edge_transform: Callable[[TupleGen], TupleGen] = None,
        global_transform: Callable[[TupleGen], TupleGen] = None,
    ):
        """Transform networkx feature data.

        .. code-block::

            def only_self_loops(edges):
                for e1, e2, edata in edges:
                    if e1 == e2:
                        yield e1, e2, edata

            transform = nx_transform(edge_transform=only_self_loops)
            transform(graphs)

        Alternatively, using the functional programming module:

        .. code-block::

            from caldera.utils.functional import Functional

            only_self_loops = Fn.filter_each(lambda x: x[0] == x[1])
            transform = nx_transform(edge_transform=only_self_loops)
            transform(graphs)

        :param node_transform:
        :param edge_transform:
        :param global_transform:
        :param kwargs:
        :return:
        """
        self.node_transform = node_transform
        self.edge_transform = edge_transform
        self.global_transform = global_transform

    def transform(self, g):
        return nx_copy(
            g,
            None,
            node_transform=self.node_transform,
            edge_transform=self.edge_transform,
            global_transform=self.global_transform,
        )


class NetworkxTransformFeatureData(NetworkxTransformFeatures):
    def __init__(
        self,
        node_transform: Callable[[Dict], Dict] = None,
        edge_transform: Callable[[Dict], Dict] = None,
        global_transform: Callable[[Dict], Dict] = None,
    ):
        def to_data(f):
            def _to_data(x):
                return tuple(list(x)[:-1]) + (f(x[-1]),)

            return _to_data

        def map_to_data(f):
            if f is None:
                return None
            return Fn.map_each(to_data(f))

        super().__init__(
            node_transform=map_to_data(node_transform),
            edge_transform=map_to_data(edge_transform),
            global_transform=map_to_data(global_transform),
        )



"""_base.py.

Transform base class
"""
from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import Generator
from typing import List
from typing import overload
from typing import Tuple
from typing import Union

import networkx as nx

from ._types import _G


class NetworkxTransformBase(ABC):
    def __repr__(self):
        return "{}".format(self.__class__.__name__)

    @abstractmethod
    def transform(self, data: _G) -> _G:
        pass

    def generate(
        self, datalist: Union[List[_G], Tuple[_G, ...], Generator[_G, None, None]]
    ) -> Generator[_G, None, None]:
        for data in datalist:
            yield self.transform(data)

    @overload
    def __call__(self, data: List[_G]) -> List[_G]:
        ...

    @overload
    def __call__(self, data: Tuple[_G]) -> Tuple[_G]:
        ...

    @overload
    def __call__(self, data: Generator[_G, None, None]) -> Generator[_G, None, None]:
        ...

    @overload
    def __call__(self, data: _G) -> _G:
        ...

    def __call__(self, data: _G) -> _G:
        if isinstance(data, list):
            transformed = list(self.generate(data))
        elif isinstance(data, tuple):
            transformed = tuple(self.generate(data))
        elif isinstance(data, nx.Graph):
            transformed = list(self.generate([data]))[0]
        else:
            transformed = self.generate(data)
        return transformed


TransformCallable = Union[Callable, NetworkxTransformBase]



from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import TypeVar
from typing import Union

T = TypeVar("T")
S = TypeVar("S")
K = TypeVar("K")
V = TypeVar("V")


def dict_join(
    a: Dict[K, T],
    b: Dict[K, S],
    out: Optional[Dict] = None,
    join_fn: Callable[[T, S], V] = None,
    default_a: T = ...,
    default_b: S = ...,
    keys: Optional[Iterable[K]] = None,
    mode: str = "union",  # Literal["union", "left", "right", "intersection"]
) -> Dict[K, Union[T, S, V]]:
    """Join two dictionaries. This function merges two dictionaries is various
    ways. For example, a dictionary of `Dict[str, List]` can be merged such
    that, if the two dictionaries share the same key, the lists are
    concatenated. The join function can be applied to the union of all keys
    (default) by (`mode="union"`), the intersection of the dictionary
    (`mode="intersection"), only the keys in the left dictionary
    (`mode="left"`), or keys only in the right dictionary (`mode="right"`).

    .. code-block::

        import operator
        d1 = {'a': [1,2,3], 'b': [1,2]
        d2 = {'a': [1,2], 'c': [10,20]
        d3 = dict_join(d1, d2, join_fn=operator.add)
        print(d3)
        # {'a': [1,2,3,1,2], 'b': [1,2], 'c': [10,20]}

    This can be done such that the first dictionary is updated instead of returning a new dictionary:

    .. code-block::

        d1 = {'a': [1,2,3], 'b': [1,2]
        d2 = {'a': [1,2], 'c': [10,20]
        dict_join(d1, d2, d2, join_fn=lambda a, b: a + b
        print(d2)
        # {'a': [1,2,3,1,2], 'b': [1,2], 'c': [10,20]}

        functools.partial(dict_join, join_fv=operator.add

    :param a: First dictionary
    :param b: Second dictionary
    :param out: The target dictionary. If None, creates a new dictionary.
    :param join_fn: Join function when both dictionaries share the same key. If not provided, will use the value
        provided by the second dictionary.
    :param default_a: Default value to use in the case a key is missing from the first dictionary. If defined as
        `Ellipsis` (or `...`), defaults will be ignored.
    :param default_b: Default value to use in the case a key is missing from the second dictionary. If defined as
        `Ellipsis` (or `...`), defaults will be ignored.
    :param keys: If provided, explicitly join only on specified keys.
    :param mode: If keys are None, mode specifies which keys to join. "union" (default), means join
        takes the union of keys from both dictionaries. "intersection" means take intersection of keys
        for both dictionaries. "left" means use only keys in the first dictionary. "right" means use only keys
        from the second dictionary.
    :return:
    """
    if out is None:
        out = dict()

    if join_fn is None:

        def join_fn(v1, v2):
            if v2 is not ...:
                return v2
            if v1 is not ...:
                return v1

    if keys is None:
        if mode == "union":
            keys = set(a).union(set(b))
        elif mode == "intersection":
            keys = set(a).intersection(set(b))
        elif mode == "left":
            keys = set(a)
        elif mode == "right":
            keys = set(b)
        else:
            raise ValueError("mode '{}' not recognized.".format(mode))

    for k in keys:
        v1 = a.get(k, default_a)
        v2 = b.get(k, default_b)
        if v1 is ...:
            v3 = v2
        elif v2 is ...:
            v3 = v1
        else:
            v3 = join_fn(v1, v2)
        out[k] = v3
    return out



r"""
.. currentmodule:: caldera.blocks

Network building blocks for creating graph neural networks.

Generic Blocks
--------------

.. autosummary::
    :toctree: generated/

    Flex
    MLP

Encoder/Decoder Blocks
----------------------

.. autosummary::
    :toctree: generated/

    NodeBlock
    EdgeBlock
    GlobalBlock

Message Passing Blocks
----------------------

.. autosummary::
    :toctree: generated/

    AggregatingNodeBlock
    AggregatingEdgeBlock
    AggregatingGlobalBlock
    Aggregator
    MultiAggregator
"""
from caldera.blocks.aggregator import Aggregator
from caldera.blocks.aggregator import MultiAggregator
from caldera.blocks.edge_block import AggregatingEdgeBlock
from caldera.blocks.edge_block import EdgeBlock
from caldera.blocks.flex import Flex
from caldera.blocks.global_block import AggregatingGlobalBlock
from caldera.blocks.global_block import GlobalBlock
from caldera.blocks.mlp import MLP
from caldera.blocks.node_block import AggregatingNodeBlock
from caldera.blocks.node_block import NodeBlock

__all__ = [
    "Aggregator",
    "MultiAggregator",
    "AggregatingEdgeBlock",
    "EdgeBlock",
    "Flex",
    "AggregatingNodeBlock",
    "AggregatingEdgeBlock",
    "AggregatingGlobalBlock",
    "NodeBlock",
    "GlobalBlock",
    "MLP",
]



import itertools
from typing import Iterable
from typing import Tuple
from typing import TypeVar

T = TypeVar("T")


def pairwise(iterable: Iterable[T]) -> Iterable[Tuple[T, T]]:
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def _first(i: Iterable[T]) -> T:
    """Select the first element in an iterable."""
    return next(x for x in itertools.tee(i)[0])



import io
from contextlib import contextmanager

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from matplotlib import cm
from PIL import Image


def edge_colors(g, key, cmap):
    edgelist = list(g.edges)
    edgefeat = list()
    for e in edgelist:
        edata = g.edges[e]
        edgefeat.append(edata[key][0].item())
    edgefeat = np.array(edgefeat)
    edgecolors = cmap(edgefeat)
    return edgecolors
    nx.draw_networkx_edges(g, pos=pos, edge_color=edgecolors, arrows=False)


def node_colors(g, key, cmap):
    nodelist = list(g.nodes)
    nodefeat = list()
    for n in nodelist:
        ndata = g.nodes[n]
        nodefeat.append(ndata[key][0].item())
    nodefeat = np.array(nodefeat)
    nodecolors = cmap(nodefeat)
    return nodecolors
    nx.draw_networkx_nodes(g, pos=pos, node_size=10, node_color=nodecolors)


def plot_graph(g, ax, cmap, key="features", seed=1):
    pos = nx.layout.spring_layout(g, seed=seed)
    nx.draw_networkx_edges(
        g, ax=ax, pos=pos, edge_color=edge_colors(g, key, cmap), arrows=False, width=1.5
    )
    nx.draw_networkx_nodes(
        g, ax=ax, pos=pos, node_size=5, node_color=node_colors(g, key, cmap)
    )


def comparison_plot(out_g, expected_g):
    fig, axes = plt.subplots(1, 2, figsize=(5, 2.5))
    axes[0].axis("off")
    axes[1].axis("off")

    axes[0].set_title("out")
    plot_graph(out_g, axes[0], cm.plasma)

    axes[1].set_title("expected")
    plot_graph(expected_g, axes[1], cm.plasma)
    return fig, axes


def validate_compare_plot(trainer, plmodel):
    eval_loader = trainer.val_dataloaders[0]
    for x, y in eval_loader:
        break
    plmodel.eval()
    y_hat = plmodel.model.forward(x, 10)[-1]
    y_graphs = y.to_networkx_list()
    y_hat_graphs = y_hat.to_networkx_list()

    idx = 0
    yg = y_graphs[idx]
    yhg = y_hat_graphs[idx]
    return comparison_plot(yhg, yg)


@contextmanager
def fig_to_pil(fig):
    buf = io.BytesIO()
    fig.savefig(buf, format="png")
    buf.seek(0)
    im = Image.open(buf)
    yield im
    buf.close()
    fig.close()


@contextmanager
def figs_to_pils(figs):
    ims = []
    buffs = []
    for fig in figs:
        buf = io.BytesIO()
        buffs.append(buf)
        fig.savefig(buf, format="png")
        buf.seek(0)
        ims.append(Image.open(buf))
    yield ims
    for buf in buffs:
        buf.close()
    plt.close()



from copy import deepcopy

from ._nx_feature_transform import NetworkxTransformFeatureData


class NetworkxDeepcopy(NetworkxTransformFeatureData):
    def __init__(self):
        """Deepcopy the graph including all of its node, edge, and global
        features.

        Typically performed before applying other transforms.
        """
        super().__init__(
            node_transform=deepcopy, edge_transform=deepcopy, global_transform=deepcopy
        )



from heapq import heappop
from heapq import heappush
from itertools import count
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import List
from typing import Optional
from typing import overload
from typing import Tuple
from typing import Union

import networkx as nx
import numpy as np

from caldera.utils.nx.traversal._path_utils import PathNpSum
from caldera.utils.nx.traversal._path_utils import PathSymbol
from caldera.utils.nx.types import Graph


def _weight_function(e, k):
    if e is None:
        return None
    return e[k]


def _multisource_dijkstra(
    g: Graph,
    sources: Union[Tuple[Hashable, ...], List[Hashable]],
    symbols: Tuple[Union[str, PathSymbol], ...],
    func: Callable = lambda x: x,
    target: Optional[Union[Tuple[Hashable, ...], List[Hashable]]] = None,
    accumulators=None,
    init=None,
    cutoff=None,
    paths=None,
    pred=None,
) -> Dict[Hashable, Union[float, int]]:

    _symbols = symbols[:]
    symbols = []
    for s in _symbols:
        if isinstance(s, str):
            s = PathSymbol(s, PathNpSum)
        symbols.append(s)

    accumulators = accumulators or {}
    init = init or {}

    # successor dictionary
    g_succ = g._succ if g.is_directed() else g._adj

    # push/pop methods to use
    push = heappush
    pop = heappop

    # node to shortest distance, breakdown
    dist_parts = {}

    # node to shortest distance
    dist = {}

    # node to shortest distance seen
    seen = {}
    c = count()
    fringe = []

    # initial/default values for each symbol

    for sym in symbols:
        if accumulators.get(sym, "sum") == "sum":
            init.setdefault(sym, 0.0)
        elif accumulators.get(sym, "product"):
            init.setdefault(sym, 1.0)
        else:
            raise ValueError("Accumulator '{}' not recognized".format(sym))
    init = np.array([s.iden_fill for s in symbols])

    # push the initial values for the sources
    for source in sources:
        if source not in g:
            raise nx.NodeNotFound("Source {} is not in G".format(source))
        seen[source] = func(*init)
        push(fringe, (0, next(c), source, init))

    # modified dijkstra's
    while fringe:
        (_, _, v, d) = pop(fringe)
        # d np.array of values
        if v in dist_parts:
            continue  # already searched this node
        dist_parts[v] = d
        dist[v] = func(*d)
        if v == target:
            break
        for u, e in g_succ[v].items():
            # vu cost breakdown for  each symbol
            costs = np.array([_weight_function(e, sym.name) for sym in symbols])

            # vu_dist break down using accumulating function
            x = np.stack([dist_parts[v], costs])
            vu_dist_parts = []
            for i, _x in enumerate(x.T):
                vu_dist_parts.append(symbols[i].accumulator(_x))
            vu_dist_parts = np.array(vu_dist_parts)
            vu_dist = func(*vu_dist_parts)

            if cutoff is not None:
                if vu_dist > cutoff:
                    continue
            if u in dist_parts:
                if vu_dist < dist[u]:
                    raise ValueError("Contradictory paths found:", "negative weights?")
            elif u not in seen or vu_dist < seen[u]:
                seen[u] = func(*vu_dist_parts)
                push(fringe, (vu_dist, next(c), u, vu_dist_parts))
                if paths is not None:
                    paths[u] = paths[v] + [u]
                if pred is not None:
                    pred[u] = v
            elif vu_dist == seen[u]:
                if pred is not None:
                    pred[u].append(v)
                    fringe = []
    return dist


@overload
def multisource_dijkstras(
    g: ...,
    sources: ...,
    symbols: ...,
    f: ...,
    target: Optional = None,
    cutoff: Optional = None,
) -> Union[Tuple[Dict[Hashable, Union[float, int]], Dict[Hashable, List[Hashable]]]]:
    ...


def multisource_dijkstras(
    g: Graph,
    sources: Union[Tuple[Hashable, ...], List[Hashable]],
    symbols: Tuple[Union[str, PathSymbol], ...],
    func: Callable = lambda x: x,
    target: Hashable = None,
    cutoff: Optional[Union[float, int]] = None,
) -> Union[Tuple[Union[float, int], List[Hashable]]]:
    if sources is None or not len(sources):
        raise ValueError("sources must not be empty")
    if target in sources:
        return (0, [target])
    paths = {source: [source] for source in sources}  # dictionary of paths
    dist = _multisource_dijkstra(
        g=g,
        sources=sources,
        func=func,
        target=target,
        symbols=symbols,
        paths=paths,
        cutoff=cutoff,
    )
    if target is None:
        return (dist, paths)
    try:
        return (dist[target], paths[target])
    except KeyError:
        raise nx.NetworkXNoPath("No _path to {}.".format(target))



import torch


class Select(torch.nn.Module):
    """Differentiable select block."""

    def __init__(self, input_size: int, output_size: int):
        super().__init__()
        self.blocks = torch.nn.Sequential(
            torch.nn.Linear(input_size, output_size), torch.nn.Sigmoid()
        )

    def forward(self, data: torch.Tensor, latent: torch.Tensor) -> torch.Tensor:
        out = self.blocks(latent)
        i = torch.where(torch.round(out) == 1)[0]
        return data[i]



import time
from typing import Union

from rich.progress import Progress


def safe_update(progress: Progress, every: Union[float, int] = 0.25):
    """Safely update a progress bar so that it displays properly in a jupyter
    notebook.

    :param progress:
    :param every: Refresh every nth second.
    :return:
    """
    last_refresh = {}

    def wrapped(task_id, *args, **kwargs):
        last_refresh.setdefault(task_id, time.time())
        dt = time.time() - last_refresh[task_id]
        if dt > every:
            last_refresh[task_id] = time.time()
            kwargs["refresh"] = True
        return progress.update(task_id, *args, **kwargs)

    return wrapped



"""Configuration utilities."""
from dataclasses import fields
from dataclasses import is_dataclass
from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

C = TypeVar("C")


def dataclass_to_dict(dataklass: C) -> Dict[str, Any]:
    """Converts a dataclass to a nested dictionary.

    :param dataklass: The dataclass instance
    :return: dictionary
    """
    data = {}
    if is_dataclass(dataklass):
        for field in fields(dataklass):
            value = getattr(dataklass, field.name)
            if is_dataclass(value):
                value = dataclass_to_dict(value)
            data[field.name] = value
    return data


def dataclass_from_dict(dataklasstype: Type[C], data: Dict[str, Any]) -> C:
    """Converts a nested dictionary to a dataclass.

    :param dataklasstype: the dataclass
    :param data: nested data as dictionary
    :return: dataclass instance
    """
    if is_dataclass(dataklasstype):
        for field in fields(dataklasstype):
            if is_dataclass(field.type):
                value = dataclass_from_dict(field.type, data[field.name])
            else:
                value = data[field.name]
            data[field.name] = value
        return dataklasstype(**data)



from itertools import tee
from typing import Callable
from typing import Generator
from typing import List
from typing import Optional
from typing import overload
from typing import Tuple
from typing import TypeVar
from typing import Union

import torch
from torch.utils.data import DataLoader
from torch.utils.data import Dataset

from caldera.data import GraphBatch
from caldera.data import GraphData
from caldera.utils import _first

T = TypeVar("T")


@overload
def collate_list(data_list: List[GraphBatch]) -> GraphBatch:
    ...


def collate_list(data_list: List[GraphData]) -> GraphBatch:
    return GraphBatch.from_data_list(data_list)


@overload
def collate_zip(data_list: List[Tuple[GraphBatch, ...]]) -> Tuple[GraphBatch, ...]:
    ...


def collate_zip(data_list: List[Tuple[GraphData, ...]]) -> Tuple[GraphBatch, ...]:
    if issubclass(type(data_list[0][0]), GraphData):
        return tuple(
            [collate_list([x[i] for x in data_list]) for i in range(len(data_list[0]))]
        )
    else:
        raise RuntimeError(
            "Cannot collate {}({})({})".format(
                type(data_list), type(data_list[0]), type(data_list[0][0])
            )
        )


class GraphDataLoader(DataLoader):
    def __init__(
        self,
        *dataset: Union[List[GraphBatch], Dataset],
        batch_size: int = 1,
        shuffle: bool = False,
        **kwargs
    ):
        if len(dataset) > 1:
            dataset = list(zip(*dataset))
            self.zipped = True
            collate = collate_zip
        else:
            dataset = dataset[0]
            collate = collate_list
        super().__init__(dataset, batch_size, shuffle, collate_fn=collate, **kwargs)

    def first(self, *args, **kwargs):
        return _first(tee(self(*args, **kwargs))[0])

    def mem_sizes(self):
        return torch.IntTensor([d.memsize() for d in self.dataset])

    def __call__(
        self,
        device: Optional[str] = None,
        f: Optional[Callable[[GraphBatch], T]] = None,
        send_to_device_before_apply: bool = True,
        limit_mem_size: Optional[int] = None,
    ) -> Generator[T, None, None]:
        """Create a new generator. Optionally apply some function to each item
        or send the item to a device. In cases where both are defined, items
        are first sent to device and then the function is applied, unless
        argument "send_to_device_before_apply" is set to False.

        :param device: optional device to send each item to
        :param f: function to apply to each item
        :param send_to_device_before_apply: if True (default), send to device before applying function (if applicable)
        :param limit_mem_size: Limit memory size of the batches. Warning, this will introduce data bias such that
            large graphs are less likely to appear together. Consider using this carefully.
        :return: generator of items
        """
        for d in self:
            if device is not None and f is not None:
                if send_to_device_before_apply:
                    ret = f(d.to(device))
                else:
                    ret = f(d).to(device)
            elif device is not None:
                ret = d.to(device)
            elif f is not None:
                ret = f(d)
            else:
                ret = d
            if limit_mem_size is not None:
                if ret.memsize() > limit_mem_size:
                    continue
            yield ret



from typing import Union

import torch

from caldera.data import GraphBatch
from caldera.data import GraphData
from caldera.data.utils._utils import edges_difference


def add_edges(data: GraphData, fill_value: Union[float, int], kind: str) -> GraphData:
    """Adds edges to the :class:`caldera.data.GraphData`.

    :param data: :class:`caldera.data.GraphData` instance.
    :param fill_value: fill value for edge attribute tensor.
    :param kind: Choose from "self" (for self edges), "complete" (for complete graph) or "undirected" (undirected edges)
    :return:
    """
    UNDIRECTED = "undirected"
    COMPLETE = "complete"
    SELF = "self"
    VALID_KIND = [UNDIRECTED, COMPLETE, SELF]
    if kind not in VALID_KIND:
        raise ValueError("'kind' must be one of {}".format(VALID_KIND))

    data_cls = data.__class__
    if issubclass(data_cls, GraphBatch):
        node_idx = data.node_idx
        edge_idx = data.edge_idx
    elif issubclass(data_cls, GraphData):
        node_idx = torch.zeros(data.x.shape[0], dtype=torch.long)
        edge_idx = torch.zeros(data.e.shape[0], dtype=torch.long)
    else:
        raise ValueError(
            "data must be a subclass of {} or {}".format(
                GraphBatch.__class__.__name__, GraphData.__class__.__name__
            )
        )

    with torch.no_grad():
        # we count the number of nodes in each graph using node_idx
        gidx, n_nodes = torch.unique(node_idx, return_counts=True, sorted=True)
        _, n_edges = torch.unique(edge_idx, return_counts=True, sorted=True)

        eidx = 0
        nidx = 0
        graph_edges_list = []
        new_edges_lengths = torch.zeros(gidx.shape[0], dtype=torch.long)

        for _gidx, _n_nodes, _n_edges in zip(gidx, n_nodes, n_edges):
            graph_edges = data.edges[:, eidx : eidx + _n_edges]
            if kind == UNDIRECTED:
                missing_edges = edges_difference(graph_edges.flip(0), graph_edges)
            elif kind == COMPLETE:
                all_graph_edges = _create_all_edges(nidx, _n_nodes)
                missing_edges = edges_difference(all_graph_edges, graph_edges)
            elif kind == SELF:
                self_edges = torch.cat(
                    [torch.arange(nidx, nidx + _n_nodes).expand(1, -1)] * 2
                )
                missing_edges = edges_difference(self_edges, graph_edges)
            graph_edges_list.append(missing_edges)

            if not missing_edges.shape[0]:
                new_edges_lengths[_gidx] = 0
            else:
                new_edges_lengths[_gidx] = missing_edges.shape[1]

            nidx += _n_nodes
            eidx += _n_edges

        new_edges = torch.cat(graph_edges_list, axis=1)
        new_edge_idx = gidx.repeat_interleave(new_edges_lengths)
        new_edge_attr = torch.full(
            (new_edges.shape[1], data.e.shape[1]), fill_value=fill_value
        )

        edges = torch.cat([data.edges, new_edges], axis=1)
        edge_idx = torch.cat([edge_idx, new_edge_idx])
        edge_attr = torch.cat([data.e, new_edge_attr], axis=0)

        idx = edge_idx.argsort()

    if issubclass(data_cls, GraphBatch):
        return GraphBatch(
            node_attr=data.x.detach().clone(),
            edge_attr=edge_attr[idx],
            global_attr=data.g.detach().clone(),
            edges=edges[:, idx],
            node_idx=data.node_idx.detach().clone(),
            edge_idx=edge_idx[idx],
        )
    elif issubclass(data_cls, GraphData):
        return data_cls(
            node_attr=data.x.detach().clone(),
            edge_attr=edge_attr[idx],
            global_attr=data.g.detach().clone(),
            edges=edges[:, idx],
        )


def _create_all_edges(start: int, n_nodes: int) -> torch.LongTensor:
    """Create all edges starting from node index 'start' for 'n_nodes'
    additional nodes.

    E.g. _all_edges(5, 4) would create all edges of nodes 5, 6, 7, 8
    with shape (2, 4)
    """
    a = torch.arange(start, start + n_nodes, dtype=torch.long)
    b = a.repeat_interleave(torch.ones(n_nodes, dtype=torch.long) * n_nodes)
    c = torch.arange(start, start + n_nodes, dtype=torch.long).repeat(n_nodes)
    return torch.stack([b, c], axis=0)



import random
from typing import overload
from typing import Tuple
from typing import Union


@overload
def _resolve_range(x: Union[Tuple[float, float], float]) -> float:
    pass


def _rand_float(a: float, b: float):
    return a + (b - a) * random.random()


@overload
def _resolve_range(x: Union[Tuple[float, float], float]) -> float:
    ...


# TODO: rename and have this choose from distributions instead
def _resolve_range(x: Union[Tuple[int, int], int]) -> int:
    if isinstance(x, int) or isinstance(x, float):
        return x
    elif isinstance(x, tuple):
        if isinstance(x[0], int):
            return random.randint(*x)
        elif isinstance(x[0], float):
            return _rand_float(*x)
    else:
        raise TypeError



from dataclasses import fields
from dataclasses import is_dataclass
from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from omegaconf import DictConfig
from omegaconf import OmegaConf


C = TypeVar("C")


def dataclass_to_dict(dataklass: C) -> Dict[str, Any]:
    """Converts a dataclass to a nested dictionary.

    :param dataklass: The dataclass instance
    :return: dictionary
    """
    data = {}
    if is_dataclass(dataklass):
        for field in fields(dataklass):
            value = getattr(dataklass, field.name)
            if is_dataclass(value):
                value = dataclass_to_dict(value)
            data[field.name] = value
    return data


def dataclass_from_dict(dataklasstype: Type[C], data: Dict[str, Any]) -> C:
    """Converts a nested dictionary to a dataclass.

    :param dataklasstype: the dataclass
    :param data: nested data as dictionary
    :return: dataclass instance
    """
    if is_dataclass(dataklasstype):
        for field in fields(dataklasstype):
            if is_dataclass(field.type):
                value = dataclass_from_dict(field.type, data[field.name])
            else:
                value = data[field.name]
            data[field.name] = value
        return dataklasstype(**data)


def validate_all_accessible(cfg: DictConfig):
    for k in cfg:
        v = cfg[k]
        if isinstance(v, DictConfig):
            validate_all_accessible(v)


def validate_structured(klass: Type, cfg: DictConfig):
    OmegaConf.structured(klass).update(cfg)


class ConfigObj:
    """Generic configuraiton object that performs validation.

    Can convert from dataclass to DictConfig.
    """

    @classmethod
    def validate_cfg(cls, cfg: DictConfig):
        validate_structured(cls, cfg)
        validate_all_accessible(cfg)

    def validate(self):
        for field in fields(self):
            value = getattr(self, field.name)
            if isinstance(value, ConfigObj):
                value.validate()

    @classmethod
    def from_dict_config(cls, cfg: DictConfig):
        cls.validate_cfg(cfg)
        return dataclass_from_dict(cls, OmegaConf.to_container(cfg, resolve=True))

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        return dataclass_from_dict(cls, data)

    def to_dict(self):
        return dataclass_to_dict(self)



import inspect
from functools import wraps
from multiprocessing import Pool
from typing import Callable
from typing import List
from typing import Optional

from ._mp_tools import _resolved_sorted_handler
from ._mp_tools import _S
from ._mp_tools import _T
from ._mp_tools import argmap
from ._mp_tools import run_with_pool
from ._mp_tools import valid_varname

# class MultiProcess(object):
#
#     def __init__(self, on: str, auto_annotate: bool = True, attach: bool = True, attach_as="pooled"):
#         if not isinstance(attach_as, str):
#             raise TypeError("attach_as must be a str, not a {}".format(attach_as.__class__))
#         if not valid_varname(attach_as):
#             raise ValueError("attach_as '{}' is not a valid variable name".format(attach_as))
#
#
#         self.on = on
#         self.auto_annotate = auto_annotate
#         self.attach = attach
#         self.attach_as = attach_as
#         self.validate()
#
#     def validate(self):
#         self.validate_on()
#         self.validate_attach_as()
#
#     def validate_attach_as(self):
#         if not isinstance(self.attach_as, str):
#             raise TypeError("attach_as must be a str, not a {}".format(self.attach_as.__class__))
#         if not valid_varname(self.attach_as):
#             raise ValueError("attach_as '{}' is not a valid variable name".format(self.attach_as))
#
#     def validate_on(self):
#         if not isinstance(self.on, str):
#             raise TypeError('argument must be a str, found a {} {}'.format(self.on, self.on.__class__))
#         elif not valid_varname(self.on):
#             raise ValueError("argument '{}' is not a valid variable name".format(self.on))
#
#     def validate_func(self, f):
#         argspec = inspect.getfullargspec(f)
#         if not self.on in argspec.args:
#             raise ValueError("argument '{}' not in signature. Use any of {}".format(self.on, argspec.args))
#
#     @staticmethod
#     def starstar(f):
#         def wrapped(kwargs):
#             return f(**kwargs)
#         return wrapped
#
#     def pooled_inner(self, f, pool_kwargs, override_kwargs_key: str):
#         @wraps(f)
#         def _pooled_inner(*args, **kwargs) -> List[_S]:
#             inner_kwargs = dict(pool_kwargs)
#             if override_kwargs_key is not None:
#                 if override_kwargs_key in kwargs:
#                     pooled_opts = kwargs[override_kwargs_key]
#                     inner_kwargs.update(pooled_opts)
#             inner_kwargs = dict(pool_kwargs)
#             n_cpus = inner_kwargs['n_cpus']
#             chunksize = inner_kwargs['chunksize']
#             callback = inner_kwargs['callback']
#             error_callback = inner_kwargs['error_callback']
#
#             amap = argmap(f, args, kwargs)
#             pooled_kwargs = []
#             if not hasattr(amap[self.on], '__iter__'):
#                 raise ValueError('argument "{}" is not iterable for pooled function'.format(on))
#             for v in amap[self.on]:
#                 _kwargs = dict(amap)
#                 _kwargs[self.on] = v
#                 pooled_kwargs.append(_kwargs)
#
#             with Pool(n_cpus) as pool:
#                 run_with_pool(pool,
#                               f,
#                               wrapper=self.starstar,
#                               chunksize=chunksize,
#                               args=pooled_kwargs,
#                               callback=callback,
#                               error_callback=error_callback,
#                               resolved_handler=_resolved_sorted_handler
#                               )
#         return _pooled_inner
#
#     # @staticmethod
#     # def override(f, outer_kwargs: Dict[str, Any], override_kwargs: str):
#     #     @wraps(f)
#     #     def wrapped(*args, **kwargs):
#     #         inner_kwargs = dict(outer_kwargs)
#     #         if override_kwargs is not None:
#     #             if override_kwargs in kwargs:
#     #                 pooled_opts = kwargs[override_kwargs]
#     #                 inner_kwargs.update(pooled_opts)
#     #         f(*args, **inner_kwargs)
#     #     return wrapped
#
#     def do_auto_annotate(self, func, type = List):
#         if self.auto_annotate:
#             func.__annotations__ = dict(func.__annotations__)
#             if self.on in func.__annotations__:
#                 func.__annotations__[self.on] = type[func.__annotations__[self.on]]
#             else:
#                 func.__annotations__[self.on] = list
#
#     def bind(self, func, pooled_outer):
#         if '__attach_as' is None or self.attach is False:
#             wrapped = pooled_outer
#         elif self.attach is True:
#             @wraps(func)
#             def wrapped(*args, **kwargs) -> _S:
#                 return func(*args, **kwargs)
#             wrapped.__dict__[self.attach_as] = pooled_outer
#         return wrapped
#
#     def override(self, f, okwargs, key):
#         def wrapped(*args, **kwargs):
#             def _wrapped(*args, **kwargs):
#                 pass
#             return _wrapped
#         return wrapped
#
#     def __call__(self, f):
#         self.validate_func(f)
#
#         def pooled_outer(n_cpus: Optional[int] = None, chunksize: int = 1, callback=None, error_callback=None,
#                          override_kwargs='pooled_opts'):
#             pool_kwargs = {
#                 'n_cpus': n_cpus,
#                 'chunksize': chunksize,
#                 'callback': callback,
#                 'error_callback': error_callback
#             }
#             pooled_inner = self.pooled_inner(f, pool_kwargs, override_kwargs)
#             if self.auto_annotate:
#                 self.do_auto_annotate(pooled_inner)
#             return pooled_inner
#
#         return self.bind(f, pooled_outer)


def _auto_annotate_on(on: str, func, type=List):
    func.__annotations__ = dict(func.__annotations__)
    if on in func.__annotations__:
        func.__annotations__[on] = type[func.__annotations__[on]]
    else:
        func.__annotations__[on] = list


def multiprocess(
    on: str, auto_annotate: bool = True, attach: bool = True, attach_as="pooled"
):
    """A decorator that attaches a pooled version of the function.

    .. code-block::

        import time

        @mp('y')
        def foo(x, y, z):
            time.sleep(x/10.)
            return x + y + z

        # standard call signature
        foo(1, 3, 2)

        # pooled call signature
        foo.pooled(1, range(20), 2)

    :param on: Variable to pool. The pooled version will require an iterable to be
               passed at the positional argument or key word argument indicated.
    :param auto_annotate: whether to dynamically change the annotation of the pooled function.
           This likely will not appear in the IDE.
    :param attach: if False, return only the pooled version of the function
    :param attach_as: key to attach to the original function
    :return:
    """
    if not isinstance(attach_as, str):
        raise TypeError("attach_as must be a str, not a {}".format(attach_as.__class__))
    if not valid_varname(attach_as):
        raise ValueError(
            "attach_as '{}' is not a valid variable name".format(attach_as)
        )

    def _mp(
        f: Callable[[_T], _S]
    ) -> Callable[[Callable[[_T], _S]], Callable[[_T], _S]]:

        if not isinstance(on, str):
            raise TypeError(
                "argument must be a str, found a {} {}".format(on, on.__class__)
            )
        elif not valid_varname(on):
            raise ValueError("argument '{}' is not a valid variable name".format(on))
        argspec = inspect.getfullargspec(f)
        if on not in argspec.args:
            raise ValueError(
                "argument '{}' not in signature. Use any of {}".format(on, argspec.args)
            )

        def pooled_outer(
            n_cpus: Optional[int] = None,
            chunksize: int = 1,
            callback=None,
            error_callback=None,
            override_kwargs="pooled_opts",
        ):

            outer_kwargs = {
                "n_cpus": n_cpus,
                "chunksize": chunksize,
                "callback": callback,
                "error_callback": error_callback,
            }

            @wraps(f)
            def pooled_inner(*args, **kwargs) -> List[_S]:
                # overrides
                inner_kwargs = dict(outer_kwargs)
                if override_kwargs is not None:
                    if override_kwargs in kwargs:
                        pooled_opts = kwargs[override_kwargs]
                        inner_kwargs.update(pooled_opts)
                inner_kwargs = dict(outer_kwargs)
                n_cpus = inner_kwargs["n_cpus"]
                chunksize = inner_kwargs["chunksize"]
                callback = inner_kwargs["callback"]
                error_callback = inner_kwargs["error_callback"]

                amap = argmap(f, args, kwargs)
                pooled_kwargs = []
                if not hasattr(amap[on], "__iter__"):
                    raise ValueError(
                        'argument "{}" is not iterable for pooled function'.format(on)
                    )
                for v in amap[on]:
                    _kwargs = dict(amap)
                    _kwargs[on] = v
                    pooled_kwargs.append(_kwargs)

                def starstar(f):
                    def wrapped(kwargs):
                        return f(**kwargs)

                    return wrapped

                with Pool(n_cpus) as pool:
                    run_with_pool(
                        pool,
                        f,
                        wrapper=starstar,
                        chunksize=chunksize,
                        args=pooled_kwargs,
                        callback=callback,
                        error_callback=error_callback,
                        resolved_handler=_resolved_sorted_handler,
                    )

            if auto_annotate:
                _auto_annotate_on(on, pooled_inner)
            return pooled_inner

        if attach_as is None or attach is False:
            wrapped = pooled_outer()
        elif attach is True:

            @wraps(f)
            def wrapped(*args, **kwargs) -> _S:
                return f(*args, **kwargs)

            wrapped.__dict__[attach_as] = pooled_outer

        return wrapped

    return _mp



"""File containing the network."""
from functools import wraps
from typing import Callable
from typing import List
from typing import Type

import torch

from .configuration import NetConfig
from caldera.blocks import AggregatingEdgeBlock
from caldera.blocks import AggregatingGlobalBlock
from caldera.blocks import AggregatingNodeBlock
from caldera.blocks import EdgeBlock
from caldera.blocks import Flex
from caldera.blocks import GlobalBlock
from caldera.blocks import MLP
from caldera.blocks import MultiAggregator
from caldera.blocks import NodeBlock
from caldera.data import GraphBatch
from caldera.models import GraphCore
from caldera.models import GraphEncoder


class EdgeBlockEncoder(EdgeBlock):
    def __init__(self, size: int, dropout: float, activation: Type[torch.nn.Module]):
        super().__init__(
            Flex(MLP)(Flex.d(), size, dropout=dropout, activation=activation)
        )


class NodeBlockEncoder(NodeBlock):
    def __init__(self, size: int, dropout: float, activation: Type[torch.nn.Module]):
        super().__init__(
            Flex(MLP)(Flex.d(), size, dropout=dropout, activation=activation)
        )


class GlobalBlockEncoder(GlobalBlock):
    def __init__(self, size: int, dropout: float, activation: Type[torch.nn.Module]):
        super().__init__(
            Flex(MLP)(Flex.d(), size, dropout=dropout, activation=activation)
        )


class NodeBlockCore(AggregatingNodeBlock):
    def __init__(
        self,
        layers,
        dropout,
        layer_norm,
        aggregator,
        aggregator_activation: Type[torch.nn.Module],
    ):
        super().__init__(
            Flex(MLP)(Flex.d(), *layers, dropout=dropout, layer_norm=layer_norm),
            Flex(MultiAggregator)(
                Flex.d(), aggregators=aggregator, activation=aggregator_activation
            ),
        )


class EdgeBlockCore(AggregatingEdgeBlock):
    def __init__(self, layers, dropout: float, layer_norm: bool):
        super().__init__(
            Flex(MLP)(Flex.d(), *layers, dropout=dropout, layer_norm=layer_norm)
        )


class GlobalBlockCore(AggregatingGlobalBlock):
    def __init__(
        self,
        layers,
        dropout,
        layer_norm,
        edge_aggregator,
        node_aggregator,
        aggregator_activation,
    ):
        super().__init__(
            mlp=Flex(MLP)(Flex.d(), *layers, dropout=dropout, layer_norm=layer_norm),
            edge_aggregator=Flex(MultiAggregator)(
                Flex.d(), aggregators=edge_aggregator, activation=aggregator_activation
            ),
            node_aggregator=Flex(MultiAggregator)(
                Flex.d(), aggregators=node_aggregator, activation=aggregator_activation
            ),
        )


class LinearTransformation(torch.nn.Module):
    def __init__(self, size: int, activation: Type[torch.nn.Module]):
        super().__init__()
        layers = [Flex(torch.nn.Linear)(Flex.d(), size)]
        if activation is not None:
            layers.append(activation())
        self.layer = torch.nn.Sequential(*layers)

    def forward(self, data):
        return self.layer(data)


# TODO: is activator getting passed?


class Network(torch.nn.Module):
    def __init__(self, config: NetConfig):
        """Do stuff with cfg.

        Use your interpreter/IDE to help construction
        """
        super().__init__()
        self.config = config
        self.encoder = self.init_encoder()
        self.core = self.init_core()
        self.decoder = self.init_encoder()
        self.output_transform = self.init_out_transform()

    # TODO: this size is going to be different, right??
    def init_encoder(self) -> GraphEncoder:
        return GraphEncoder(
            edge_block=EdgeBlockEncoder(
                self.config.encode.edge.size,
                self.config.encode.edge.dropout,
                self.config.get_activation(self.config.encode.edge.activation),
            ),
            node_block=NodeBlockEncoder(
                self.config.encode.node.size,
                self.config.encode.node.dropout,
                self.config.get_activation(self.config.encode.node.activation),
            ),
            global_block=GlobalBlockEncoder(
                self.config.encode.glob.size,
                self.config.encode.glob.dropout,
                self.config.get_activation(self.config.encode.glob.activation),
            ),
        )

    # def init_decoder(self) -> GraphEncoder:
    #     return GraphEncoder(
    #         edge_block=EdgeBlockEncoder(
    #             self.config.decode.edge.size,
    #             self.config.decode.edge.dropout,
    #             self.config.get_activation(self.config.decode.edge.activation),
    #         ),
    #         node_block=NodeBlockEncoder(
    #             self.config.decode.node.size,
    #             self.config.decode.node.dropout,
    #             self.config.get_activation(self.config.decode.node.activation),
    #         ),
    #         global_block=GlobalBlockEncoder(
    #             self.config.decode.glob.size,
    #             self.config.decode.glob.dropout,
    #             self.config.get_activation(self.config.decode.glob.activation),
    #         ),
    #     )

    def init_core(self):
        return GraphCore(
            edge_block=EdgeBlockCore(
                layers=self.config.core.edge.layers,
                dropout=self.config.core.edge.dropout,
                layer_norm=self.config.core.edge.layer_norm,
            ),
            node_block=NodeBlockCore(
                layers=self.config.core.node.layers,
                dropout=self.config.core.node.dropout,
                aggregator=self.config.edge_block_to_node_aggregators,
                aggregator_activation=self.config.get_activation(
                    self.config.aggregator_activation
                ),
                layer_norm=self.config.core.node.layer_norm,
            ),
            global_block=GlobalBlockCore(
                layers=self.config.core.glob.layers,
                dropout=self.config.core.glob.dropout,
                edge_aggregator=self.config.global_block_to_edge_aggregators,
                node_aggregator=self.config.global_block_to_node_aggregators,
                layer_norm=self.config.core.glob.layer_norm,
                aggregator_activation=self.config.get_activation(
                    self.config.aggregator_activation
                ),
            ),
            pass_global_to_node=self.config.pass_global_to_node,
            pass_global_to_edge=self.config.pass_global_to_edge,
        )

    # TODO: each edge should have its own activation
    def init_out_transform(self):
        return GraphEncoder(
            edge_block=EdgeBlock(
                LinearTransformation(
                    self.config.out.edge.size,
                    self.config.get_activation(self.config.out.edge.activation),
                ),
            ),
            node_block=NodeBlock(
                LinearTransformation(
                    self.config.out.node.size,
                    self.config.get_activation(self.config.out.node.activation),
                )
            ),
            global_block=GlobalBlock(
                LinearTransformation(
                    self.config.out.glob.size,
                    self.config.get_activation(self.config.out.glob.activation),
                )
            ),
        )

    # TODO: fix the ordering of e, x, g and graph tuple
    def encode(self, data: GraphBatch) -> GraphBatch:
        e, x, g = self.encoder(data)
        return GraphBatch(x, e, g, data.edges, data.node_idx, data.edge_idx)

    def decode(self, data: GraphBatch) -> GraphBatch:
        e, x, g = self.decoder(data)
        return GraphBatch(x, e, g, data.edges, data.node_idx, data.edge_idx)

    def core_process(self, latent0: GraphBatch, data: GraphBatch) -> GraphBatch:
        e = torch.cat([latent0.e, data.e], dim=1)
        x = torch.cat([latent0.x, data.x], dim=1)
        g = torch.cat([latent0.g, data.g], dim=1)
        data = GraphBatch(x, e, g, data.edges, data.node_idx, data.edge_idx)
        e, x, g = self.core(data)
        return GraphBatch(x, e, g, data.edges, data.node_idx, data.edge_idx)

    def out_transform(self, data: GraphBatch) -> GraphBatch:
        e, x, g = self.output_transform(data)
        return GraphBatch(x, e, g, data.edges, data.node_idx, data.edge_idx)

    def forward(
        self, data: GraphBatch, steps: int, save_all: bool = False
    ) -> List[GraphBatch]:
        data = self.encode(data)
        latent0 = data

        outputs = []
        for _ in range(steps):
            data = self.core_process(latent0, data)
            data = self.decode(data)
            out_data = self.out_transform(data)
            if save_all:
                outputs.append(out_data)
            else:
                outputs = [out_data]
        return outputs

    @wraps(forward)
    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)



r"""
Networkx utilities.

.. autosummary::
    :toctree: generated/

    nx_copy
    nx_copy_to_directed
    nx_copy_to_undirected
    nx_deep_copy
    nx_is_directed
    nx_is_undirected
    nx_iter_leaves
    nx_iter_roots
    nx_shallow_copy
    nx_to_directed
    nx_to_undirected

Traversals
----------

.. autosummary::
    :toctree: generated/

   traversal.floyd_warshall
   traversal.PathAccumulator
   traversal.PathMax
   traversal.PathMin
   traversal.PathMul
   traversal.PathNpProduct
   traversal.PathNpSum
   traversal.PathSum
   traversal.PathSymbol
   traversal.multisource_dijkstras

Generators
----------

.. autosummary::
    :toctree: generated/

    generators.chain_graph
    generators.compose_and_connect
    generators.connect_node_sets
    generators.nx_random_features
    generators.rand_n_nodes_n_edges
    generators.random_edge
    generators.random_node
    generators.random_graph
    generators.unique_chain_graph
"""
from ._tools import nx_copy
from ._tools import nx_copy_to_directed
from ._tools import nx_copy_to_undirected
from ._tools import nx_deep_copy
from ._tools import nx_is_directed
from ._tools import nx_is_undirected
from ._tools import nx_iter_leaves
from ._tools import nx_iter_roots
from ._tools import nx_shallow_copy
from ._tools import nx_to_directed
from ._tools import nx_to_undirected

__all__ = [
    "nx_copy",
    "nx_copy_to_undirected",
    "nx_copy_to_directed",
    "nx_deep_copy",
    "nx_is_directed",
    "nx_is_undirected",
    "nx_iter_leaves",
    "nx_iter_roots",
    "nx_shallow_copy",
    "nx_to_directed",
    "nx_to_undirected",
]



from typing import Optional
from typing import Union

import torch

from caldera.data import GraphBatch
from caldera.data import GraphData
from caldera.utils.sparse import scatter_coo


def to_sparse_coo_matrix(
    data: Union[GraphData, GraphBatch],
    fill_value: Optional[Union[int, float, torch.Tensor]] = None,
    dtype=torch.float,
):
    """Return the sparse coo representation of the edge attribute data of the.

    :class:`caldera.data.GraphData` instance. If provided with the optional
    fill value, will return a sparse matrix filled with the value instead of
    the edge attribute data.

    .. code-block::

        to_sparse_coo_matrix(data)
        to_sparse_coo_matrix(data, fill_value=1)

    :param data:
    :param fill_value:
    :param dtype:
    :return:
    """
    if fill_value is not None:
        return scatter_coo(
            data.edges,
            fill_value,
            expand=True,
            dtype=dtype,
            size=(data.num_nodes, data.num_nodes, ...),
        )
    else:
        return scatter_coo(
            data.edges,
            data.e,
            expand=False,
            dtype=dtype,
            size=(data.num_nodes, data.num_nodes, ...),
        )



r"""
Caldera models

.. autosummary::
    :toctree: generated/

    GraphEncoder
    GraphCore
    EncodeCoreDecode
"""
from caldera.models.encoder_core_decoder import EncodeCoreDecode
from caldera.models.graph_core import GraphCore
from caldera.models.graph_encoder import GraphEncoder

__all__ = ["EncodeCoreDecode", "GraphCore", "GraphEncoder"]



from ._nx_apply_to_key import NetworkxApplyToKey


class NetworkxFlattenNodeFeature(NetworkxApplyToKey):
    def __init__(self, key: str):
        super().__init__(key, node_func=lambda x: x.flatten())


class NetworkxFlattenEdgeFeature(NetworkxApplyToKey):
    def __init__(self, key: str):
        super().__init__(key, edge_func=lambda x: x.flatten())


class NetworkxFlattenGlobalFeature(NetworkxApplyToKey):
    def __init__(self, key: str):
        super().__init__(key, glob_func=lambda x: x.flatten())



"""_random_hop.py.

Choose random nodes, and induce the graph on those nodes to "h" hopes.
"""
from typing import Dict
from typing import Optional
from typing import overload

import torch

from ._base import TransformBase
from caldera.data import GraphBatch
from caldera.data import GraphData
from caldera.data.utils import induce


class RandomHop(TransformBase):
    def __init__(self, n_nodes: int, n_hops: int):
        """Choose subgraphs from a graph.

        :param n_nodes:
        :param n_hops:
        """
        self.n_nodes = n_nodes
        self.n_hops = n_hops

    @overload
    def transform_each(self, data: GraphData, edge_dict: ...) -> GraphData:
        ...

    def transform(
        self, data: GraphBatch, edge_dict: Optional[Dict] = None
    ) -> GraphBatch:
        if data.num_nodes == 0:
            return data.clone()
        src = torch.randint(data.num_nodes, (self.n_nodes,))
        nodes = induce(data, src, self.n_hops, edge_dict=edge_dict)
        node_mask = torch.BoolTensor([False] * data.num_nodes)
        node_mask[nodes] = True
        return data.apply_node_mask(node_mask)



"""functional.py.

Functional programming methods.

Generally, functions are named by `verb-noun-qualifier`

`verbs`

**ignore** - returned function ignores certain elements of the iterator, returning an iterator of the remaining
elements.
**return** - returns the result of the function, rather than another iterator.
**group** - returned function returns an iterator of groups (i.e. iterables of each element `T`, such as `List[T]`),
instead of
**map** - applies a new function to elements of the iterator

`nouns`

**each** - returned function applies to each element in the iterator
**all** - returned function applies to the entire iterator, rather than to each element

`qualifier`

**until** - iterates until some condition is met

**Special Methods (i.e. piping methods)**

`compose`
"""
import functools
import itertools
import random
from collections import OrderedDict
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

T = TypeVar("T")
S = TypeVar("S")
U = TypeVar("U")

GT = Generator[T, None, None]
GS = Generator[S, None, None]


def _validate_positive_int(n: int):
    if n is not None:
        if not isinstance(n, int):
            raise TypeError(
                "`n` must be an integer greater than 0 or None, not {}".format(n)
            )
        elif n < 0:
            raise ValueError(
                "`n` must be an integer greater than 0 or None, not {}".format(n)
            )


def compose(*funcs: Tuple[Callable[[GT], GS], ...]) -> Callable[[GT], GS]:
    """Create a function composition.

    :param funcs:
    :return:
    """

    def _compose(data):
        result = data
        for i, f in enumerate(funcs):

            def _composition_part(r):
                try:
                    return f(result)
                except Exception as e:
                    raise e
                    msg = str(e)
                    msg += "\nduring composition:"
                    msg += "\n ({}) f: {}".format(i, f)
                    msg += "\n args: {} {}".format(result, result.__class__)
                    raise type(e)(msg) from e

            result = _composition_part(result)
        return result

    return _compose


def map_each(f: Callable[[T], S], *args, **kwargs) -> Callable[[GS], GT]:
    def _map_each(arr):
        for a in arr:
            yield f(a, *args, **kwargs)

    return _map_each


def map_all(f: Callable[[GT], GS], *args, **kwargs) -> Callable[[GT], GS]:
    """Apply the function to the iterator itself.

    .. note::     Of course, this could be done just as easily by
    supplying the provided     function itself. This keeps the syntax
    common with other methods.
    """

    def _map_all(arr: GT) -> GS:
        return f(arr, *args, **kwargs)

    return _map_all


def map_all_if(condition, if_func, else_func=None):
    def _map_all_if(arr):
        for a in arr:
            if condition(a):
                yield if_func(a)
            elif else_func is not None:
                yield else_func(a)
            else:
                yield a

    return _map_all_if


def zip_each_with(*arrs, first: bool = False):
    """.. warning::

        If provided with generators or generator functions,
        calling this function will consume all the provided iterables during
        iteration. The provided generators should never
        be used again elsewhere.

    :param arrs:
    :return:
    """

    def _zip_each_with(a):
        if first:
            yield from zip(a, *arrs)
        else:
            yield from zip(*arrs, a)

    return _zip_each_with


def cat(*arrs):
    """.. warning::

        If provided with generators or generator functions,
        calling this function will consume all the provided iterables
        immediately. The provided generators should never
        be used again elsewhere.

    :param arrs:
    :return:
    """
    arrs = [list(a) for a in arrs]

    def _cat(a):
        yield a
        yield from arrs

    return _cat


def zipmap_each_with(*funcs, first: bool = True):
    def _zipmap_each_with(arr):
        yield from compose(
            zip_each_with(funcs, first=first),
            map_each(lambda x: x[1](x[0])),
        )(arr)

    return _zipmap_each_with


def iter_each_unique():
    def _iter_each_unique(arr):
        seen = set()
        for a in arr:
            if a not in seen:
                seen.add(a)
                yield a

    return _iter_each_unique


def iter_reverse():
    def _iter_reverse(arr):
        yield from iter(list(arr)[::-1])

    return _iter_reverse


def zip_all(reverse: bool = False):
    def _zip_all(arr):
        if reverse:
            yield from zip(*list(arr)[::-1])
        else:
            yield from zip(*arr)

    return _zip_all


def reduce_each(f):
    def reducer(arr):
        return functools.reduce(f, arr)

    def _reduce_each(arr):
        yield from map_each(reducer)(arr)

    return _reduce_each


def reduce_all(f):
    def reducer(arr):
        return functools.reduce(f, arr)

    def _reduce_all(arr):
        yield from map_all(reducer)(arr)

    return _reduce_all

    # iter_map_items = functools.partial(map_each, zipped=True)


def group_each_consecutive(condition: Callable[[T], bool]):
    def _group_each_consecutive(arr):
        chunk = []
        v = ...
        for a in arr:
            new_v = condition(a)
            if v is ... or new_v == v:
                chunk.append(a)
            else:
                yield chunk
                chunk = [a]
            v = new_v
        if len(chunk):
            yield chunk

    return _group_each_consecutive


def group_each_until(
    condition: Callable[[List[T]], bool],
    group_if: Optional[Callable[[T], bool]] = None,
    yield_final: bool = True,
) -> Callable[[GT], Generator[List[T], None, None]]:
    def _group_each_until(arr):
        chunk = []
        for a in arr:
            if group_if:
                if group_if(a):
                    chunk.append(a)
            else:
                chunk.append(a)

            if condition(chunk):
                yield chunk
                chunk = []

        if yield_final or condition(chunk):
            yield chunk

    return _group_each_until


def group_each_into_chunks(
    chunk_size: int,
) -> Callable[[Iterable[T]], Generator[T, None, None]]:
    """Returns a new function the groups iterables by chunks of the specified
    size. Last chunk is returned by default.

    .. doctest::

        >>> from caldera.utils.functional import group_each_into_chunks
        >> a = [1, 2, 3, 4]
        >> f = fn_chunks(3)
        >> print(f(a))
        [[1, 2, 3], [4]]

    :param chunk_size: size of each chunk
    :return: generator of chunks
    """

    def _group_each_into_chunks(arr):
        yield from group_each_until(
            lambda g: len(g) >= chunk_size and len(g), yield_final=True
        )(arr)

    return _group_each_into_chunks


def repeat_all(n: Optional[int]) -> Callable[[GT], GT]:
    if n is not None:
        _validate_positive_int(n)

    def _repeat_all(arr: Iterable[T]) -> Generator[T, None, None]:
        c = itertools.tee(arr, 1)[0]
        i = 0
        while n is None or i < n:
            b, c = itertools.tee(c)
            yield from b
            i += 1

    return _repeat_all


def counter(i: Optional[int] = 0) -> Generator[int, None, None]:
    while True:
        yield i
        i += 1


def enumerate_each(*, reverse: bool = False):
    """Non-tee version of enumerate."""

    def _enumerate_each(arr):
        if not reverse:
            for i, a in enumerate(arr):
                yield i, a
        else:
            for i, a in enumerate(arr):
                yield a, i

    return _enumerate_each


def iter_count(n: int = 1):
    if n == 0:
        return empty_generator
    _validate_positive_int(n)

    def _iter_count(arr):
        yield from compose(
            enumerate_each(),
            until(lambda x: x[0] < n - 1, yield_last=True),
            index_each(1),
        )(arr)

    return _iter_count


def iter_next():
    def _iter_count(arr):
        yield from iter_count(1)(arr)

    return _iter_count


def consume(f: Callable[[GT], Any]) -> Callable[[GT], GT]:
    """Applies a function to the iterator and then consumes.

    :param f:
    :return:
    """

    def _consume(arr):
        list(f(arr))
        return arr

    return _consume


def empty_generator(*_):
    return
    yield


def tee_pipe(*funcs: Tuple[Callable[[GT], GS], ...]) -> Callable[[GT], GT]:
    """Tee the iterator and apply a new series of functions without affecting
    the primary iterator.

    .. code-blocK:

        def iter_print(arr):
            for a in arr:
                print(a)
                yield a

        func = functional.pipe(
            functional.side_effect(iter_print, functional.map_each(lambda x: x**2), iter_print),
            functional.map_each(lambda x: x**2)
        )

    :return:
    """

    piped = compose(*funcs)

    def _tee_pipe(arr):
        a, b = itertools.tee(arr)
        piped(a)
        return b

    return _tee_pipe


def tee_pipe_yield(*funcs):
    piped = compose(*funcs)

    def _tee_pipe_yield(arr):
        a, b = itertools.tee(arr)
        yield piped(a)
        yield b

    return _tee_pipe_yield


def tee_consume(*funcs):
    piped = compose(*funcs)

    def _tee_consume(arr):
        list(piped(arr))
        return arr

    return _tee_consume


def apply_each(func):
    def _apply_each(arr):
        for a in arr:
            func(a)
            yield a

    return _apply_each


def index_each(i: int) -> Callable[[GT], GS]:
    def _index_each(arr):
        yield from map_each(lambda x: x[i])(arr)

    return _index_each


def index_from(d: Dict) -> Callable[[GT], GS]:
    def _index_from(arr):
        yield from map_each(lambda x: d[x])(arr)

    return _index_from


def get_each(i: int, default=...) -> Callable[[GT], GT]:
    def _get_each(arr):
        if default is ...:
            yield from map_each(lambda x: x.get(i))(arr)
        else:
            yield from map_each(lambda x: x.get(i, default))(arr)

    return _get_each


def getattr_each(k: str):
    def _getattr_each(arr):
        yield from map_each(lambda x: getattr(x, k))(arr)

    return _getattr_each


def get_from(d: Dict, default=...) -> Callable[[GT], GT]:
    def _get_from(arr):
        if default is ...:
            yield from map_each(lambda x: d.get(x))(arr)
        else:
            yield from map_each(lambda x: d.get(x, default))(arr)

    return _get_from


def _reverse(x):
    if isinstance(x, tuple):
        return tuple(list(x)[::-1])
    elif isinstance(x, list):
        return x[::-1]
    else:
        return list(x)[::-1]


def reverse_each():
    def _reverse_each(arr):
        for a in arr:
            yield _reverse(a)

    return _reverse_each


def ignore_each_count(
    n: int = 1,
) -> Callable[[Iterable[T]], Generator[T, None, None]]:
    """pipe( fn_side_effect(fn_next(n)) )

    :param n:
    :return:
    """

    def _ignore_each_count(arr: Iterable[T]) -> Generator[T, None, None]:
        for i, a in enumerate(arr):
            if i >= n:
                yield a

    return _ignore_each_count


def yield_all(*funcs: Tuple[Callable, ...]):
    def _yield_all(arr):
        if not funcs:
            yield arr
        else:
            for _f in funcs:
                yield _f(arr)

    return _yield_all


def ignore_each_until(
    f: Callable[[T], bool]
) -> Callable[[Iterable[T]], Generator[T, None, None]]:
    def _ignore_each_until(arr: Iterable[T]) -> Generator[T, None, None]:
        start = False
        for a in arr:
            if not start and f(a):
                start = True
            if start:
                yield a

    return _ignore_each_until


def iter_each_until(
    f: Callable[[T], bool]
) -> Callable[[Iterable[T]], Generator[T, None, None]]:
    def _iter_each_until(arr: Iterable[T]) -> Generator[T, None, None]:
        for a in arr:
            if f(a):
                yield a
                return
            yield a

    return _iter_each_until


def iter_step(
    step_size: int = 1,
) -> Callable[[Iterable[T]], Generator[T, None, None]]:
    _validate_positive_int(step_size)

    def _iter_step(arr):
        for i, a in enumerate(arr):
            if i % step_size == 0:
                yield a

    return _iter_step


def tee_all(n: int = 2):
    def _tee_all(arr: Iterable[T]) -> Generator[T, None, None]:
        yield from itertools.tee(arr, n)

    return _tee_all


def trycatch(*exceptions, catch_yields=...):
    """Catches the provided exceptions. If one of the exceptions is raised,
    will stop iteration. If provided with `catch_yields`, the value will be
    yielded upon exception.

    :param exceptions:
    :param catch_yields:
    :return:
    """

    def _trycath(arr):
        while True:
            try:
                yield next(arr)
            except Exception as e:
                for valid_exception in exceptions:
                    if isinstance(e, valid_exception):
                        if catch_yields is not ...:
                            yield catch_yields
                        return
                raise e

    return _trycath


def pairwise_each() -> Callable[[GT], Generator[Tuple[T, T], None, None]]:
    def _pairwise_each(arr: GT) -> Generator[Tuple[T, T], None, None]:
        a, b = itertools.tee(arr)
        next(b)
        return zip(a, b)

    return _pairwise_each


def filter_each(f: Callable[[T], bool], inverse: bool = False) -> Callable[[GT], GT]:
    def _filter_each(arr: GT) -> GT:
        for a in arr:
            v = f(a)
            if not inverse and v:
                yield a
            elif inverse and not v:
                yield a

    return _filter_each


def until(
    f: Callable[[T], bool], inverse: bool = False, yield_last: bool = False
) -> Callable[[GT], GT]:
    def _until(arr: GT) -> GT:
        for a in arr:
            v = f(a)
            if not inverse and v:
                yield a
            elif inverse and not v:
                yield a
            else:
                if yield_last:
                    yield a
                return

    return _until


def chain_each() -> Callable[[Generator[GT, None, None]], GT]:
    def _chain_each(arr: Generator[GT, None, None]) -> GT:
        for a in arr:
            yield from a

    return _chain_each


def group_each_by_key(
    f: Callable[[T], S], value_func: Optional[Callable[[S], U]] = None
) -> Callable[[GT], Generator[Tuple[S, Union[T, U]], None, None]]:
    def _group_each_by_key(arr):
        a, b = itertools.tee(arr)
        data = OrderedDict()
        for _a in a:
            k = f(_a)
            data.setdefault(k, list())
            if value_func:
                _a = value_func(_a)
            data[k].append(_a)
        yield from data.items()

    return _group_each_by_key


def star(f):
    """Return a function equivalent to `lambda args: f(*args)`

    :param f:
    :return:
    """

    def _star(arr):
        return f(*arr)

    return _star


def count_each():
    def _count_each(arr):
        for i, arr in enumerate(arr):
            yield i

    return _count_each


def apply(func):
    def _apply(arr):
        func(arr)
        return arr

    return _apply


def shuffle_each():
    def _shuffle_each(arr):
        shuffled = random.shuffle(arr)
        yield from shuffled

    return _shuffle_each



import os
from multiprocessing import Pool
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar

import networkx as nx
import numpy as np
from rich.progress import Progress

from .configuration import DataConfig
from .configuration import DataGenConfig
from .progress import safe_update
from caldera.data import GraphData
from caldera.data import GraphDataLoader
from caldera.testing import annotate_shortest_path
from caldera.transforms import Compose
from caldera.transforms.networkx import NetworkxAttachNumpyBool
from caldera.transforms.networkx import NetworkxAttachNumpyOneHot
from caldera.transforms.networkx import NetworkxNodesToStr
from caldera.transforms.networkx import NetworkxSetDefaultFeature
from caldera.transforms.networkx import NetworkxToDirected
from caldera.utils.mp import multiprocess
from caldera.utils.nx.generators import chain_graph
from caldera.utils.nx.generators import compose_and_connect
from caldera.utils.nx.generators import random_graph
from caldera.utils.nx.generators import uuid_sequence

T = TypeVar("T")


def generate_shortest_path_example(n_nodes, density, path_length, compose_density=None):
    d = float(density)
    l = int(path_length)
    if compose_density is None:
        cd = d
    else:
        cd = float(compose_density)
    path = list(uuid_sequence(l))
    h = chain_graph(path, nx.Graph)
    g = random_graph(n_nodes, density=d)
    graph = compose_and_connect(g, h, cd)

    annotate_shortest_path(
        graph,
        True,
        True,
        source_key="source",
        target_key="target",
        path_key="shortest_path",
        source=path[0],
        target=path[-1],
    )

    preprocess = Compose(
        [
            NetworkxSetDefaultFeature(
                node_default={"source": False, "target": False, "shortest_path": False},
                edge_default={"shortest_path": False},
            ),
            NetworkxAttachNumpyOneHot(
                "node", "source", "_features", classes=[False, True]
            ),
            NetworkxAttachNumpyOneHot(
                "node", "target", "_features", classes=[False, True]
            ),
            NetworkxAttachNumpyOneHot(
                "edge", "shortest_path", "_target", classes=[False, True]
            ),
            NetworkxAttachNumpyOneHot(
                "node", "shortest_path", "_target", classes=[False, True]
            ),
            NetworkxSetDefaultFeature(
                node_default={
                    "_features": np.array([1.0, 0.0]),
                    "_target": np.array([1.0, 0.0]),
                },
                edge_default={
                    "_features": np.array([1.0, 0.0]),
                    "_target": np.array([1.0, 0.0]),
                },
                global_default={
                    "_features": np.array([1.0]),
                    "_target": np.array([1.0]),
                },
            ),
            NetworkxNodesToStr(),
            NetworkxToDirected(),
        ]
    )

    return preprocess([graph])[0]


def _star_generate_shortest_path_example(args):
    return generate_shortest_path_example(*args)


@multiprocess(on="graph")
def graph_to_data(graph, key):
    return GraphData.from_networkx(graph, feature_key=key)


class DataGenerator:
    def __init__(self, config: DataConfig, progress_bar: bool = True):
        self.config = config
        self._train_loader = None
        self._eval_loader = None
        self.progress_bar = progress_bar

    def _create_nxg(self, data_config):
        return generate_shortest_path_example(
            data_config.num_nodes,
            data_config.density,
            data_config.path_length,
            data_config.composition_density,
        )

    def _create_raw_data(
        self,
        data_config: DataGenConfig,
        desc: Optional[str] = None,
        callback: Callable = None,
    ) -> List[nx.DiGraph]:
        graphs = []

        def new_arg():
            return (
                data_config.num_nodes,
                data_config.density,
                data_config.path_length,
                data_config.composition_density,
            )

        def new_args(n):
            return [new_arg() for _ in range(n)]

        args = new_args(data_config.num_graphs)

        with Pool() as pool:
            results = pool.imap_unordered(
                _star_generate_shortest_path_example, args, chunksize=100
            )

            if desc is None:
                desc = "generating {} data ({} cpus)".format(
                    data_config.name, os.cpu_count()
                )
            while True:
                try:
                    result = next(results)
                    graphs.append(result)
                    if callback:
                        callback(result)
                except StopIteration:
                    break

        return graphs

    def _create_dataloader(
        self,
        input_feature_key: str,
        target_feature_key: str,
        config: DataGenConfig,
        progress_bar: Optional[bool] = None,
    ):
        progress_bar = progress_bar or self.progress_bar
        with Progress(auto_refresh=True, refresh_per_second=1) as progress:

            task1 = progress.add_task(
                "[red]Generating {} data...".format(config.name),
                total=config.num_graphs,
                visible=progress_bar,
            )
            task2 = progress.add_task(
                "[purple]Converting {} data...".format(config.name),
                total=config.num_graphs,
                visible=progress_bar,
            )

            update = safe_update(progress, every=0.1)

            graphs = self._create_raw_data(
                config, callback=lambda _: update(task1, advance=1, refresh=False)
            )
            input_datalist = []
            target_datalist = []

            for graph in graphs:
                input_datalist.append(
                    GraphData.from_networkx(graph, feature_key=input_feature_key)
                )
                update(task2, advance=0.5, refresh=False)
                target_datalist.append(
                    GraphData.from_networkx(graph, feature_key=target_feature_key)
                )
                update(task2, advance=0.5, refresh=False)

        loader = GraphDataLoader(
            input_datalist,
            target_datalist,
            batch_size=config.batch_size,
            shuffle=config.shuffle,
        )
        return loader

    def _create_train_loader(self):
        return self._create_dataloader("_features", "_target", self.config.train)

    def _create_eval_loader(self):
        return self._create_dataloader("_features", "_target", self.config.eval)

    def train_loader(self) -> GraphDataLoader:
        if self._train_loader is None:
            self._train_loader = self._create_train_loader()
        return self._train_loader

    def eval_loader(self) -> GraphDataLoader:
        if self._eval_loader is None:
            self._eval_loader = self._create_eval_loader()
        return self._eval_loader

    def reset(self) -> None:
        self._train_loader = None
        self._eval_loader = None

    def init(self) -> Tuple[GraphDataLoader, GraphDataLoader]:
        self.train_loader()
        self.eval_loader()
        return self



r"""

Caldera utility functions.

.. autosummary::
    :toctree: generated/

    dict_join
    # pairwise


Indexing
--------

.. autosummary::
    :toctree: generated/

    reindex_tensor
    unravel_index

Tensor
------

Utilities for :class:`torch.Tensor`

.. autosummary::
    :toctree: generated/

    scatter_coo
    scatter_indices
    torch_coo_to_scipy_coo
    deterministic_seed
    long_isin
    same_storage
    stable_arg_sort_long
    tensor_is_empty
    torch_scatter_group

Functional
----------

Functional programming module.

.. autosummary::
    :toctree: generated/
    :recursive:

    functional

Networkx Utilities
------------------

Extra :mod:`networkx` utilities

.. autosummary::
    :toctree: generated/
    :recursive:

"""
from ._dict_join import dict_join
from ._iteration import _first
from ._iteration import pairwise
from caldera.utils.indexing import reindex_tensor
from caldera.utils.indexing import unravel_index
from caldera.utils.np import replace_nan_with_inf
from caldera.utils.sparse import scatter_coo
from caldera.utils.sparse import scatter_indices
from caldera.utils.sparse import torch_coo_to_scipy_coo
from caldera.utils.tensor import deterministic_seed
from caldera.utils.tensor import long_isin
from caldera.utils.tensor import same_storage
from caldera.utils.tensor import stable_arg_sort_long
from caldera.utils.tensor import tensor_is_empty
from caldera.utils.tensor import torch_scatter_group

__all__ = [
    "reindex_tensor",
    "unravel_index",
    "scatter_coo",
    "scatter_indices",
    "torch_coo_to_scipy_coo",
    "deterministic_seed",
    "long_isin",
    "same_storage",
    "stable_arg_sort_long",
    "tensor_is_empty",
    "torch_scatter_group",
    "dict_join",
    "pairwise",
    "_first",
    "replace_nan_with_inf",
]



import random
from typing import List
from typing import overload
from typing import Tuple

import numpy
import torch


def tensor_is_empty(x: torch.Tensor) -> bool:
    """Return whether the tensor is empty."""
    return 0 in x.shape


def same_storage(
    x: torch.Tensor, y: torch.Tensor, empty_does_not_share_storage: bool = True
) -> bool:
    """Checks if two tensors share storage.

    :param x: first tensor
    :param y: second tensor
    :param empty_does_not_share_storage: if True (default), will return False if
        either tensor is empty (despite that they technically data_ptr are the same).
    :return: if the tensor shares the same storage
    """
    if empty_does_not_share_storage and (tensor_is_empty(x) or tensor_is_empty(y)):
        return False
    x_ptrs = {e.data_ptr() for e in x.flatten()}
    y_ptrs = {e.data_ptr() for e in y.flatten()}
    return (x_ptrs <= y_ptrs) or (y_ptrs <= x_ptrs)


def deterministic_seed(seed: int, cudnn_deterministic: bool = False):
    random.seed(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    if cudnn_deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


@overload
def to_one_hot(arr: numpy.ndarray, mx: int) -> numpy.ndarray:
    ...


def to_one_hot(arr: torch.tensor, mx: int) -> torch.tensor:
    if torch.is_tensor(arr):

        oh = torch.zeros((arr.shape[0], mx))
    else:
        oh = numpy.zeros((arr.shape[0], mx))
    for i, a in enumerate(arr):
        oh[i, a] = 1.0
    return oh


def stable_arg_sort(arr, mn: float, mx: float):
    dim = -1
    if not dim == -1:
        raise ValueError("only last dimension sort is supported. Try reshaping tensor.")
    delta_shape = list(arr.shape)
    delta_shape[dim] = 1
    delta = torch.linspace(mn, mx, arr.shape[dim], device=arr.device)
    delta = delta.repeat(delta_shape)
    return torch.argsort(arr + delta, dim=dim)


def stable_arg_sort_long(arr):
    """Stable sort of long tensors.

    Note that Pytorch 1.5.0 does not have a stable sort implementation.
    Here we simply add a delta value between 0 and 1 (exclusive) and
    assuming we are using integers, call torch.argsort to get a stable
    sort.
    """
    if not (arr.dtype == torch.long or arr.dtype == torch.int):
        raise ValueError("only torch.Long or torch.Int allowed")
    return stable_arg_sort(arr, 0.0, 0.99)


def torch_scatter_group(
    x: torch.Tensor, idx: torch.Tensor
) -> Tuple[torch.Tensor, List[torch.Tensor]]:
    """Group a tensor by indices. This is equivalent to successive applications
    of `x[torch.where(x == index)]` for all provided sorted indices.

    Example:

    .. code-block::

        idx = torch.tensor([2, 2, 0, 1, 1, 1, 2])
        x = torch.tensor([0, 1, 2, 3, 4, 5, 6])

        uniq_sorted_idx, out = scatter_group(x, idx)

        # node the idx is sorted
        assert torch.all(torch.eq(out[0], torch.tensor([0, 1, 2])))

        # where idx == 0
        assert torch.all(torch.eq(out[1][0], torch.tensor([2])))

        # where idx == 1
        assert torch.all(torch.eq(out[1][1], torch.tensor([3, 4, 5])))

        # where idx == 2
        assert torch.all(torch.eq(out[1][2], torch.tensor([0, 1, 6])))

    :param x: tensor to group
    :param idx: indices
    :return: tuple of unique, sorted indices and a list of tensors corresponding to the groups
    """
    arg = stable_arg_sort_long(idx)
    x = x[arg]
    groups, b = torch.unique(idx, return_counts=True)
    i_a = 0
    arr_list = []
    for i_b in b:
        arr_list.append(x[i_a : i_a + i_b.item()])
        i_a += i_b.item()
    return groups, arr_list


def long_isin(ar1, ar2, assume_unique: bool = False, invert: bool = False):
    dim = -1
    if ar1.dtype != torch.long or ar2.dtype != torch.long:
        raise ValueError("Arrays be torch.LongTensor")
    if ar2.ndim > 1:
        raise ValueError(
            "Unable to broadcast shape {}. Second tensor must be a "
            "1-dimensional.".format(ar2.shape)
        )

    # Otherwise use sorting
    if not assume_unique:
        ar1, rev_idx = torch.unique(ar1, return_inverse=True)
        ar2 = torch.unique(ar2, dim=None)
        # TODO: how to handle repeats and unique in multidimensional tensor?

    # if ar2.ndim > 1:
    #     s = list(ar2.shape)
    #     s[dim] = 1
    #     ar1 = ar1.repeat(s)
    ar = torch.cat((ar1, ar2), axis=dim)

    # We need this to be a stable sort
    order = stable_arg_sort_long(ar)
    sar = torch.gather(ar, dim, order)
    if invert:
        bool_ar = sar[1:] != sar[:-1]
    else:
        bool_ar = sar[1:] == sar[:-1]
    flag = torch.cat((bool_ar, torch.tensor([invert])))
    ret = torch.empty(ar.shape, dtype=bool)
    ret[order] = flag

    if assume_unique:
        return ret[: len(ar1)]
    else:
        return ret[rev_idx]



r"""
Methods for transforming :class:`networkx.Graph`s

.. currentmodule:: caldera.transforms.networkx

.. warning::

    You should assume all transforms are performed in place. Use
    :class:`NetworkxDeepcopy` if a copy of the graph is wanted.

.. autosummary::
    :toctree: generated/

    NetworkxApplyToFeature
    NetworkxAttachNumpyFeatures
    NetworkxAttachNumpyOneHot
    NetworkxAttachNumpyBool
    NetworkxSetDefaultFeature
    NetworkxTransformFeatureData
    NetworkxTransformFeatures
    NetworkxFlattenEdgeFeature
    NetworkxFlattenGlobalFeature
    NetworkxFlattenNodeFeature
    NetworkxNodesToStr
    NetworkxToDirected
    NetworkxToUndirected
    NetworkxFilterDataKeys
    NetworkxDeepcpy
"""
from ._nx_apply_to_graph import NetworkxApply
from ._nx_apply_to_key import NetworkxApplyToKey
from ._nx_attach_np_features import NetworkxAttachNumpyBool
from ._nx_attach_np_features import NetworkxAttachNumpyFeatures
from ._nx_attach_np_features import NetworkxAttachNumpyOneHot
from ._nx_deepcopy import NetworkxDeepcopy
from ._nx_defaults import NetworkxSetDefaultFeature
from ._nx_feature_transform import NetworkxTransformFeatureData
from ._nx_feature_transform import NetworkxTransformFeatures
from ._nx_filter_keys import NetworkxFilterDataKeys
from ._nx_flatten_feature import NetworkxFlattenEdgeFeature
from ._nx_flatten_feature import NetworkxFlattenGlobalFeature
from ._nx_flatten_feature import NetworkxFlattenNodeFeature
from ._nx_nodes_to_str import NetworkxNodesToStr
from ._nx_to_directed import NetworkxToDirected
from ._nx_to_undirected import NetworkxToUndirected


__all__ = [
    "NetworkxAttachNumpyFeatures",
    "NetworkxAttachNumpyOneHot",
    "NetworkxAttachNumpyBool",
    "NetworkxSetDefaultFeature",
    "NetworkxTransformFeatureData",
    "NetworkxTransformFeatures",
    "NetworkxNodesToStr",
    "NetworkxToDirected",
    "NetworkxToUndirected",
    "NetworkxApplyToKey",
    "NetworkxFlattenNodeFeature",
    "NetworkxFlattenEdgeFeature",
    "NetworkxFlattenGlobalFeature",
    "NetworkxFilterDataKeys",
    "NetworkxApply",
    "NetworkxDeepcopy",
]



from typing import Callable

import networkx as nx
import numpy as np


class PathAccumulator:

    all = {}

    def __init__(
        self,
        name: str,
        func: Callable[[np.ndarray, np.ndarray], np.ndarray],
        i_fill: float,
    ):
        self.name = name
        self.func = func
        if self.name in self.all:
            raise ValueError("{} already defined".format(name))
        self.all[self.name] = self
        self.i_fill = i_fill

    def __call__(self, *args, **kwargs) -> np.ndarray:
        return self.func(*args, **kwargs)

    @classmethod
    def get(cls, name, default: str = None):
        if default is None:
            return cls.all[name]
        else:
            if name not in cls.all:
                name = default
            return cls.all.get(name)


PathMul = PathAccumulator("product", np.multiply, 1.0)
PathSum = PathAccumulator("sum", lambda a, b: a + b, 0.0)
PathMin = PathAccumulator("min", np.minimum, 0.0)
PathMax = PathAccumulator("max", np.maximum, 0.0)
PathNpSum = PathAccumulator("npsum", np.sum, 0.0)
PathNpProduct = PathAccumulator("npprod", np.product, 1.0)


class PathSymbol:
    def __init__(
        self,
        name: str,
        accumulator: Callable = PathSum,
        nonedge_fill=np.inf,
        multigraph_weight: Callable = sum,
    ):
        self.name = name
        self.iden_fill = accumulator.i_fill
        self.multigraph_weight = multigraph_weight
        self.nonedge_fill = nonedge_fill
        self.accumulator = accumulator

    def initialize_matrix(self, g, dtype, nodelist=None):
        if nodelist is None:
            nodelist = list(g.nodes)

        W = nx.to_numpy_matrix(
            g,
            nodelist=nodelist,
            multigraph_weight=self.multigraph_weight,
            weight=self.name,
            nonedge=self.nonedge_fill,
            dtype=dtype,
        )

        n, m = W.shape

        # replace diagonals
        identity = np.identity(n)

        # set accumulators
        W[identity == 1] = self.accumulator.i_fill
        return W



import torch

from caldera.blocks import AggregatingEdgeBlock
from caldera.blocks import AggregatingGlobalBlock
from caldera.blocks import AggregatingNodeBlock
from caldera.blocks import Aggregator
from caldera.blocks import EdgeBlock
from caldera.blocks import Flex
from caldera.blocks import GlobalBlock
from caldera.blocks import MLP
from caldera.blocks import NodeBlock
from caldera.data import GraphBatch
from caldera.models.graph_core import GraphCore
from caldera.models.graph_encoder import GraphEncoder


class EncodeCoreDecode(torch.nn.Module):
    def __init__(
        self,
        latent_sizes=(128, 128, 1),
        output_sizes=(1, 1, 1),
        depths=(1, 1, 1),
        layer_norm: bool = True,
        dropout: float = None,
        pass_global_to_edge: bool = True,
        pass_global_to_node: bool = True,
    ):
        super().__init__()
        self.config = {
            "latent_size": {
                "node": latent_sizes[1],
                "edge": latent_sizes[0],
                "global": latent_sizes[2],
                "core_node_block_depth": depths[0],
                "core_edge_block_depth": depths[1],
                "core_global_block_depth": depths[2],
            },
            "output_size": {
                "edge": output_sizes[0],
                "node": output_sizes[1],
                "global": output_sizes[2],
            },
            "node_block_aggregator": "add",
            "global_block_to_node_aggregator": "add",
            "global_block_to_edge_aggregator": "add",
            "pass_global_to_edge": pass_global_to_edge,
            "pass_global_to_node": pass_global_to_node,
        }

        def mlp(*layer_sizes):
            return Flex(MLP)(
                Flex.d(), *layer_sizes, layer_norm=layer_norm, dropout=dropout
            )

        self.encoder = GraphEncoder(
            EdgeBlock(mlp(latent_sizes[0])),
            NodeBlock(mlp(latent_sizes[1])),
            GlobalBlock(mlp(latent_sizes[2])),
        )

        edge_layers = [self.config["latent_size"]["edge"]] * self.config["latent_size"][
            "core_edge_block_depth"
        ]
        node_layers = [self.config["latent_size"]["node"]] * self.config["latent_size"][
            "core_node_block_depth"
        ]
        global_layers = [self.config["latent_size"]["global"]] * self.config[
            "latent_size"
        ]["core_global_block_depth"]

        self.core = GraphCore(
            AggregatingEdgeBlock(mlp(*edge_layers)),
            AggregatingNodeBlock(
                mlp(*node_layers), Aggregator(self.config["node_block_aggregator"])
            ),
            AggregatingGlobalBlock(
                mlp(*global_layers),
                edge_aggregator=Aggregator(
                    self.config["global_block_to_edge_aggregator"]
                ),
                node_aggregator=Aggregator(
                    self.config["global_block_to_node_aggregator"]
                ),
            ),
            pass_global_to_edge=self.config["pass_global_to_edge"],
            pass_global_to_node=self.config["pass_global_to_node"],
        )

        self.decoder = GraphEncoder(
            EdgeBlock(mlp(latent_sizes[0])),
            NodeBlock(mlp(latent_sizes[1])),
            GlobalBlock(mlp(latent_sizes[2])),
        )

        self.output_transform = GraphEncoder(
            EdgeBlock(Flex(torch.nn.Linear)(Flex.d(), output_sizes[0])),
            NodeBlock(Flex(torch.nn.Linear)(Flex.d(), output_sizes[1])),
            GlobalBlock(Flex(torch.nn.Linear)(Flex.d(), output_sizes[2])),
        )

    def forward(self, data, steps, save_all: bool = True):
        # encoded
        e, x, g = self.encoder(data)
        data = GraphBatch(x, e, g, data.edges, data.node_idx, data.edge_idx)

        # graph topography data
        edges = data.edges
        node_idx = data.node_idx
        edge_idx = data.edge_idx
        latent0 = data

        meta = (edges, node_idx, edge_idx)

        outputs = []
        for _ in range(steps):
            # core processing step
            e = torch.cat([latent0.e, e], dim=1)
            x = torch.cat([latent0.x, x], dim=1)
            g = torch.cat([latent0.g, g], dim=1)
            data = GraphBatch(x, e, g, *meta)
            e, x, g = self.core(data)

            # decode
            data = GraphBatch(x, e, g, *meta)

            _e, _x, _g = self.decoder(data)
            decoded = GraphBatch(_x, _e, _g, *meta)

            # transform
            _e, _x, _g = self.output_transform(decoded)
            gt = GraphBatch(_x, _e, _g, edges, node_idx, edge_idx)
            if save_all:
                outputs.append(gt)
            else:
                outputs[0] = gt
        return outputs



from collections import OrderedDict
from typing import Callable
from typing import Hashable
from typing import List
from typing import Type

import numpy as np

from caldera.utils.np import replace_nan_with_inf
from caldera.utils.nx.traversal._path_utils import PathSum
from caldera.utils.nx.traversal._path_utils import PathSymbol
from caldera.utils.nx.types import Graph


def floyd_warshall(
    g: Graph,
    symbols: List[PathSymbol],
    func: Callable = None,
    nodelist: List[Hashable] = None,
    return_all: bool = False,
    dtype: Type = np.float64,
):
    """Run the floyd-warshall algorithm (all pairs shortest _path) with
    arbitrary cost functions.

    .. code-block::

        W = floyd_warshall2(g, symbols=[
                PathSymbol("A", SumPath),
                PathSymbol("B", MulPath)
            ], func: lambda a, b: a / b
        )

    .. code-block::

        W = floyd_warshall2(g, key="weight")

    :param g:
    :param symbols:
    :param func:
    :param nodelist:
    :param return_all:
    :param dtype:
    :return:
    """

    _symbols = symbols[:]
    symbols = []
    for s in _symbols:
        if isinstance(s, str):
            s = PathSymbol(s, PathSum)
        symbols.append(s)

    if func is None:

        def func(x):
            return x

    if nodelist is None:
        nodelist = list(g.nodes())

    # initialize weight matrices
    matrix_dict = OrderedDict()
    for symbol in symbols:
        matrix_dict[symbol.name] = symbol.initialize_matrix(g, dtype, nodelist)

    if return_all:
        ori_matrix_dict = {k: v.copy() for k, v in matrix_dict.items()}

    n, m = list(matrix_dict.values())[0].shape

    for i in np.arange(n):
        # get costs if using node 'i'
        parts_dict = OrderedDict()
        for symbol in symbols:
            M = matrix_dict[symbol.name]
            parts_dict[symbol.name] = symbol.accumulator(M[i, :], M[:, i])

        # get total cost
        m_arr = [np.asarray(m) for m in matrix_dict.values()]
        p_arr = [np.asarray(m) for m in parts_dict.values()]
        C = replace_nan_with_inf(func(*m_arr))
        C_part = func(*p_arr)

        # update
        for key, M in matrix_dict.items():
            part = parts_dict[key]
            c = C > C_part

            if np.any(c):
                # assert M.shape == part.shape
                np.putmask(M, c, part)

    m_arr = [np.asarray(m) for m in matrix_dict.values()]
    C = replace_nan_with_inf(func(*m_arr))
    if return_all:
        return C, matrix_dict, ori_matrix_dict
    return C



from .config import Config
from .config import DataConfig
from .config import get_config
from .config import HyperParamConfig
from .config import NetConfig
from .data import DataGenConfig

__all__ = ["Config", "NetConfig", "DataConfig", "HyperParamConfig", "get_config"]



r"""
Graphdata sets

.. currentmodule:: caldera.dataset

.. autosummary::
    :toctree: generated/

    GraphDataset
    GraphBatchDataset
"""
from ._graph_dataset import GraphBatchDataset
from ._graph_dataset import GraphDataset

__all__ = ["GraphDataset", "GraphBatchDataset"]



import itertools
import random
from copy import deepcopy as do_deepcopy
from typing import Callable
from typing import Dict
from typing import Generator
from typing import Hashable
from typing import Optional
from typing import overload
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union

import networkx as nx

from caldera.utils.nx._global_accessor import GraphWithGlobal


T = TypeVar("S")


@overload
def _resolve_range(x: Union[Tuple[float, float], float]) -> float:
    pass


def _rand_float(a: float, b: float):
    return a + (b - a) * random.random()


@overload
def _resolve_range(x: Union[Tuple[float, float], float]) -> float:
    ...


def _resolve_range(x: Union[Tuple[int, int], int]) -> int:
    if isinstance(x, int) or isinstance(x, float):
        return x
    elif isinstance(x, tuple):
        if isinstance(x[0], int):
            return random.randint(*x)
        elif isinstance(x[0], float):
            return _rand_float(*x)
    else:
        raise TypeError


def nx_iter_roots(g: nx.DiGraph) -> Generator[Hashable, None, None]:
    for n in g.nodes():
        if not list(g.predecessors(n)):
            yield n


def nx_iter_leaves(g: nx.DiGraph) -> Generator[Hashable, None, None]:
    for n in g.nodes():
        if not list(g.successors(n)):
            yield n


NodeGenerator = Generator[Tuple[T, Dict], None, None]
EdgeGenerator = Generator[Tuple[T, T, Dict], None, None]


def nx_copy(
    from_graph: nx.Graph,
    to_graph: Union[nx.Graph, Type[nx.Graph]],
    *,
    node_transform: Optional[Callable[[NodeGenerator], NodeGenerator]] = None,
    edge_transform: Optional[Callable[[EdgeGenerator], EdgeGenerator]] = None,
    global_transform: Optional[Callable[[NodeGenerator], NodeGenerator]] = None,
    global_attr_key: str = None,
    deepcopy: bool = False
) -> nx.Graph:
    """Copies node, edges, node_data, and edge_data from graph `g1` to graph
    `g2`. If `g2` is None, a new graph of type `g1.__class__` is created. If
    `g2` is a class or subclass of `nx.Graph`, a new graph of that type is
    created.

    If `deepcopy` is set to True, the copy will perform a deepcopy of all node
    and edge data. If false, only shallow copies will be created.

    `node_transform` and `edge_transform` can be provided to perform a transform
    on the on `g1.nodes(data=True)` and `g1.edges(data=True)` iterators
    during copy. The node transform should return a `Generator[Tuple[T, dict], None, None]`
    while the edge transform should return a `Generator[Tuple[T, T, dict], None, None]`.
    These transforms may include skipping certain nodes or edges, transforming the node or edge
    data, or transforming the node keys themselves.

    Some example transforms include:

    .. code-block::

        def node_to_str(gen):
            for n, ndata in gen:
                yield (str(n), ndata)

        def remove_self_loops(gen):
            for n1, n2, edata in gen:
                if n1 == n2:
                    yield (n1, n2, edata)

        nx_copy(g1, None, node_transform=node_to_str, edge_transform=remove_self_loops, deepcopy=True)


    :param from_graph: graph to copy from
    :param to_graph: graph to copy to
    :param node_transform: optional transform applied to the `from_graph.nodes(data=True)` iterator
    :param edge_transform: optional transform applied to the `from_graph.edges(data=True)` iterator
    :param literal_transform: if True, node_transform will *not* be applied following the edge_transform.
        This may result in unintentional edges being created.
    :param deepcopy:
    :return:
    """
    if to_graph is None:
        to_graph = from_graph.__class__()
    elif isinstance(to_graph, type) and issubclass(to_graph, nx.Graph):
        to_graph = to_graph()

    node_iter = from_graph.nodes(data=True)
    if node_transform:
        niter1, niter2 = itertools.tee(node_iter)
        node_iter = list(node_transform(niter1))
        _node_mapping = {}
        for x1, x2 in zip(niter2, node_iter):
            _node_mapping[x1[0]] = x2[0]

        def map_node(x):
            return _node_mapping[x]

    else:

        def map_node(x):
            return x

    for n, ndata in node_iter:
        if deepcopy:
            n, ndata = do_deepcopy((n, ndata))
        to_graph.add_node(n, **ndata)

    edge_iter = from_graph.edges(data=True)
    if edge_transform:
        edge_iter = edge_transform(edge_iter)

    for n1, n2, edata in edge_iter:
        if deepcopy:
            n1, n2, edata = do_deepcopy((n1, n2, edata))
        to_graph.add_edge(map_node(n1), map_node(n2), **edata)

    if hasattr(from_graph, GraphWithGlobal.get_global.__name__) and hasattr(
        to_graph, GraphWithGlobal.get_global.__name__
    ):
        giter = from_graph.globals(data=True, global_key=global_attr_key)
        if global_transform:
            giter = global_transform(giter)
        for gkey, gdata in giter:
            if deepcopy:
                gdata = do_deepcopy(gdata)
            else:
                gdata = dict(gdata)
            to_graph.set_global(gdata, gkey)
    return to_graph


def nx_shallow_copy(g1: nx.Graph, g2: Optional[nx.Graph] = None) -> nx.Graph:
    return nx_copy(g1, g2, deepcopy=False)


def nx_deep_copy(g1: nx.Graph, g2: Optional[nx.Graph] = None) -> nx.Graph:
    return nx_copy(g1, g2, deepcopy=True)


def nx_class_is_undirected(cls: Type):
    return cls in [nx.Graph, nx.OrderedGraph, nx.MultiGraph, nx.OrderedMultiGraph]


def nx_is_undirected(g: nx.Graph) -> bool:
    return nx_class_is_undirected(g.__class__)


def nx_class_is_directed(cls: Type):
    return cls in [
        nx.DiGraph,
        nx.OrderedDiGraph,
        nx.MultiDiGraph,
        nx.OrderedMultiDiGraph,
    ]


def nx_is_directed(g: nx.Graph) -> bool:
    return nx_class_is_directed(g.__class__)


def nx_copy_to_undirected(
    g: nx.DiGraph, graph_class: Type[nx.Graph] = nx.Graph
) -> Type[nx.Graph]:
    if not nx_class_is_undirected(graph_class):
        raise TypeError(
            "graph_class must be a directed graph type, but found {}".format(
                graph_class
            )
        )
    new_graph = graph_class()
    return nx_deep_copy(g, new_graph)


def nx_to_undirected(
    g: nx.DiGraph, graph_class: Type[nx.Graph] = nx.Graph
) -> Type[nx.Graph]:
    if not nx_class_is_undirected(graph_class):
        raise TypeError(
            "graph_class must be a directed graph type, but found {}".format(
                graph_class
            )
        )
    new_graph = graph_class()
    return nx_shallow_copy(g, new_graph)


def _nx_to_directed(g):
    for n1, n2, edata in g.edges(data=True):
        if not g.has_edge(n2, n1):
            g.add_edge(n2, n1, **edata)


def nx_to_directed(
    g: nx.Graph, graph_class: Type[nx.DiGraph] = nx.DiGraph
) -> Type[nx.Graph]:
    if not nx_class_is_directed(graph_class):
        raise TypeError(
            "graph_class must be a directed graph type, but found {}".format(
                graph_class
            )
        )
    copied = graph_class()
    nx_shallow_copy(g, copied)
    if nx_is_undirected(g):
        _nx_to_directed(copied)
    return copied


def nx_copy_to_directed(
    g: nx.Graph, graph_class: Type[nx.DiGraph] = nx.DiGraph
) -> Type[nx.Graph]:
    if not nx_class_is_directed(graph_class):
        raise TypeError(
            "graph_class must be a directed graph type, but found {}".format(
                graph_class
            )
        )
    copied = graph_class()
    nx_deep_copy(g, copied)
    if nx_is_undirected(copied):
        _nx_to_directed(copied)
    return copied



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['caldera',
 'caldera._setup',
 'caldera.blocks',
 'caldera.data',
 'caldera.data.utils',
 'caldera.dataset',
 'caldera.models',
 'caldera.testing',
 'caldera.transforms',
 'caldera.transforms.networkx',
 'caldera.utils',
 'caldera.utils.functional',
 'caldera.utils.mp',
 'caldera.utils.nx',
 'caldera.utils.nx.generators',
 'caldera.utils.nx.traversal',
 'examples',
 'examples.archived',
 'examples.crn',
 'examples.crn.visualization',
 'examples.extras',
 'examples.traversals',
 'examples.traversals.configuration']

package_data = \
{'': ['*'],
 'examples': ['sigmoid_circuit/*', 'tour/*'],
 'examples.crn': ['wandb/run-20200902_070851-2zfxdt4c/*'],
 'examples.traversals': ['conf/*',
                         'conf/data/*',
                         'conf/hyperparameters/*',
                         'conf/logger/*',
                         'conf/network/*',
                         'conf/settings/help/*',
                         'conf/settings/logging/*']}

install_requires = \
['attrs>=20.2.0,<21.0.0',
 'colorama>=0.4.3,<0.5.0',
 'dill>=0.3.2,<0.4.0',
 'networkx>=2.4,<3.0',
 'numpy>=1.18.4,<2.0.0',
 'pytest-randomly>=3.4.1,<4.0.0',
 'scipy>=1.5.2,<2.0.0',
 'torch-scatter>=2.0.4,<3.0.0',
 'torch>=1.5.0,<2.0.0']

extras_require = \
{'docs': ['sphinx_autodoc_typehints>=1.11.0,<2.0.0',
          'sphinx>=3.2.1,<4.0.0',
          'keats>=0.2.28,<0.3.0',
          'sphinx_bootstrap_theme>=0.7.1,<0.8.0',
          'nbformat>=5.0.7,<6.0.0',
          'nbconvert>=5.6.1,<6.0.0',
          'jupyter-sphinx>=0.3.1,<0.4.0'],
 'lint': ['pytest-cov>=2.10.0,<3.0.0',
          'black>=19.10b0,<20.0',
          'pre-commit>=2.6.0,<3.0.0',
          'pylint>=2.5.3,<3.0.0'],
 'nbexamples': ['seaborn>=0.11.0,<0.12.0',
                'matplotlib>=3.3.1,<4.0.0',
                'tqdm>=4.48.2,<5.0.0',
                'pytorch-lightning>=0.9.0,<0.10.0',
                'hydra-core>=1.0.0,<2.0.0',
                'rich>=7.0.0,<8.0.0',
                'h5py>=2.10.0,<3.0.0'],
 'xtests': ['pytest-sugar>=0.9.4,<0.10.0']}

setup_kwargs = {
    'name': 'pytorch-caldera',
    'version': '0.1.0a0',
    'description': '',
    'long_description': None,
    'author': 'Justin Vrana',
    'author_email': 'justin.vrana@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://www.github.com/jvrana/caldera',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os
from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Union

from hydra.conf import ConfigStore
from hydra.experimental import compose as hydra_compose_config
from hydra.experimental import initialize_config_dir
from omegaconf import DictConfig
from omegaconf import OmegaConf
from py.path import local

from .data import DataConfig
from .data import DiscreteUniform
from .data import Uniform
from .hyperparameters import HyperParamConfig
from .network import NetConfig
from .tools import ConfigObj


@dataclass
class Config(ConfigObj):
    """Configuration dataclass.

    .. note::

        To access properties from a omegaconf.DictConfig instance,
        call the following

        .. code-block::

            from caldera.utils.conf import dataclass_from_dict, dataclass_to_dict

            config = dataclass_from_dict(Config, dict(cfg))
    """

    network: NetConfig = field(default_factory=NetConfig)
    hyperparameters: HyperParamConfig = field(default_factory=HyperParamConfig)
    data: DataConfig = field(default_factory=DataConfig)
    gpus: int = 1
    log_level: str = "WARNING"
    offline: bool = False


def initialize_config():
    #     """Initialize structured configuration."""
    cs = ConfigStore()
    cs.store(name="config", node=Config)
    cs.store(group="network", name="default", node=NetConfig)
    cs.store(group="hyperparameters", name="default", node=HyperParamConfig)
    cs.store(group="data", name="default", node=DataConfig)


# TODO: make this a metaclass and automatic?
def initialize_resolvers():
    OmegaConf.register_resolver(Uniform.name, Uniform)
    OmegaConf.register_resolver(DiscreteUniform.name, DiscreteUniform)


def get_config(
    overrides: List[str] = None,
    config_path: str = "conf",
    config_name: str = "config",
    directory: str = None,
    as_config_class: bool = False,
) -> Union[DictConfig, Config]:
    """Get config (instead of running command line, as in a jupyter notebook.

    :param overrides: list of config overrides
    :param config_path: config directory path
    :param config_name: main config name
    :param directory:
    :return: DictConfig configuration
    """
    initialize_config()
    directory = directory or os.getcwd()
    with local(directory).as_cwd():
        overrides = overrides or []
        config_path = os.path.join(directory, config_path)
        with initialize_config_dir(config_path):
            cfg = hydra_compose_config(config_name=config_name, overrides=overrides)
    if as_config_class:
        cfg = Config.from_dict_config(cfg)
    return cfg


initialize_config()
initialize_resolvers()



import logging

from rich.console import Console
from rich.logging import RichHandler


# logging.basicConfig(
#     level="DEBUG",
#     format="%(message)s",
#     datefmt="[%X]",
#     handlers=[RichHandler(markup=True, show_time=False)]
# )

console = Console()
logger = logging.getLogger("examples.traversals")
logger.addHandler(RichHandler(console=console, markup=True, show_time=False))



"""_reverse.py.

Reverse the direction of the edges.
"""
from typing import overload

from ._base import TransformBase
from caldera.data import GraphBatch
from caldera.data import GraphData


class Reverse(TransformBase):
    @overload
    def transform_each(self, data: GraphData) -> GraphData:
        ...

    def transform(self, data: GraphBatch) -> GraphBatch:
        return data.reverse()



from typing import Dict
from typing import Hashable
from typing import List
from typing import Optional
from typing import overload
from typing import Set
from typing import Tuple
from typing import Union

import torch

from ._floyd_warshall import floyd_warshall
from ._utils import get_edge_dict
from caldera.data import GraphBatch
from caldera.data import GraphData
from caldera.utils import long_isin
from caldera.utils import torch_scatter_group


@overload
def tensor_induce(data: ..., nodes: torch.BoolTensor, k: ...) -> torch.BoolTensor:
    ...


def tensor_induce(
    data: Union[GraphData, GraphBatch], nodes: torch.LongTensor, k: int
) -> torch.LongTensor:
    if isinstance(nodes, int):
        nodes = torch.LongTensor([nodes])
    elif nodes.dtype == torch.long and nodes.ndim == 0:
        nodes = nodes.expand(1)

    visited = nodes.detach().clone()
    for _k in range(k):
        nodes = neighbors(data, nodes)
        if nodes.dtype == torch.bool:
            visited = torch.logical_or(visited, nodes)
            if visited.sum() >= data.num_nodes:
                break
        else:
            visited = torch.unique(torch.cat([visited, nodes]), sorted=True)
            if visited.shape[0] >= data.num_nodes:
                break
    return visited


@overload
def induce(data: ..., nodes: torch.BoolTensor, k: ...) -> torch.BoolTensor:
    ...


def induce(
    data: Union[GraphData, GraphBatch],
    nodes: torch.LongTensor,
    k: int,
    edge_dict: Optional[Dict] = None,
) -> torch.LongTensor:
    assert nodes.ndim == 1
    if nodes.dtype == torch.long:
        visited = bfs_nodes(nodes, data.edges, depth=k, edge_dict=edge_dict)
        ret = torch.tensor(list(visited), dtype=torch.long)
        return ret
    elif nodes.dtype == torch.bool:
        nidx = torch.where(nodes)[0]
        visited = bfs_nodes(nidx, data.edges, depth=k, edge_dict=edge_dict)
        ret = torch.tensor([False] * data.num_nodes)
        ret[torch.LongTensor(list(visited))] = True
        return ret
    else:
        raise ValueError("{} is not a valid type".format(data.dtype))


@overload
def neighbors(data: ..., nodes: torch.BoolTensor) -> torch.BoolTensor:
    ...


# TODO: dispatch method with multidim neighbors and floyd-warshall
def neighbors(
    data: Union[GraphData, GraphBatch],
    nodes: torch.LongTensor,
    reverse: bool = False,
    undirected: bool = False,
) -> torch.LongTensor:
    """Return the neighbors of the provided nodes.

    :param data:
    :param nodes:
    :return:
    """
    if isinstance(nodes, int):
        nodes = torch.LongTensor([nodes])
    elif nodes.dtype == torch.long and nodes.ndim == 0:
        nodes = nodes.expand(1)
    is_bool = False
    if nodes.dtype == torch.bool:
        is_bool = True
        nodes = torch.where(nodes)[0]

    if undirected:
        reachable1 = long_isin(data.edges[0], nodes)
        dest1 = data.edges[1][reachable1]
        reachable2 = long_isin(data.edges[1], nodes)
        dest2 = data.edges[0][reachable2]
        dest = torch.unique(torch.cat([dest1, dest2]))
    else:
        if reverse:
            i, j = 1, 0
        else:
            i, j = 0, 1
        reachable = long_isin(data.edges[i], nodes)
        dest = data.edges[j][reachable]

    if is_bool:
        ret = torch.full((data.num_nodes,), False, dtype=torch.bool)
        ret[dest] = True
    else:
        ret = torch.unique(dest, sorted=True)
    return ret


def bfs_nodes(
    src: Union[int, List[int], Tuple[int, ...], torch.LongTensor],
    edges: torch.LongTensor,
    depth: Optional[int] = None,
    edge_dict: Optional[Dict] = None,
) -> Set[Hashable]:
    """Return nodes from a breadth-first search. Optionally provide a depth.

    :param src:
    :param edges:
    :param depth:
    :return:
    """
    if edge_dict is None:
        edge_dict = get_edge_dict(edges)
    if torch.is_tensor(src):
        nlist = src.tolist()
    elif isinstance(src, list):
        nlist = src[:]
    elif isinstance(src, tuple):
        nlist = list(src)
    elif isinstance(src, int):
        nlist = [src]

    to_visit = nlist[:]
    depths = [0] * len(nlist)
    discovered = set()

    i = 0
    while to_visit and (depth is None or i < depth):
        v = to_visit.pop(0)
        d = depths.pop(0)
        if depth is not None and d > depth:
            continue

        discovered.add(v)
        if depth is None or d + 1 <= depth:
            if v in edge_dict:
                neighbors = edge_dict[v]
                for n in neighbors:
                    if n not in discovered:
                        to_visit.append(n)
                        depths.append(d + 1)
    return discovered


def _fw_neighbors(m, n, depth):
    x = m[n]
    reachable = torch.logical_and(x != float("inf"), x <= depth)
    if n.dtype == torch.bool:
        return reachable
    idx = torch.where(reachable)
    return idx


@overload
def floyd_warshall_neighbors(
    data: ...,
    nodes: Tuple[torch.LongTensor, ...],
    depth: ...,
    matrix: ...,
    return_matrix: ...,
) -> Union[torch.LongTensor, Tuple[torch.LongTensor, torch.FloatTensor]]:
    ...


@overload
def floyd_warshall_neighbors(
    data: ...,
    nodes: Tuple[torch.BoolTensor, ...],
    depth: ...,
    matrix: ...,
    return_matrix: ...,
) -> Union[torch.BoolTensor, Tuple[torch.BoolTensor, torch.FloatTensor]]:
    ...


def floyd_warshall_neighbors(
    data: Union[GraphData, GraphBatch],
    nodes: Union[torch.LongTensor, int],
    depth: int = 1,
    matrix: Optional[torch.FloatTensor] = None,
    return_matrix: bool = False,
) -> Union[torch.LongTensor, Tuple[torch.LongTensor, torch.FloatTensor]]:
    if matrix is None:
        matrix = floyd_warshall(data)

    if isinstance(nodes, int):
        nodes = torch.LongTensor(nodes)

    if not isinstance(nodes, tuple) and not nodes.ndim <= 2:
        raise ValueError("`nodes` must be a 0, 1, or 2-dimensional tensor")

    if isinstance(nodes, tuple):
        ret = []
        for n in nodes:
            ret.append(floyd_warshall_neighbors(data, n, depth=depth, matrix=matrix))
        ret = tuple(ret)
    else:
        idx = _fw_neighbors(matrix, nodes, depth)
        if torch.is_tensor(nodes) and nodes.ndim == 2:
            ret = torch_scatter_group(idx[-1], idx[0])
        else:
            ret = idx[-1]

    if return_matrix:
        ret = (ret, matrix)
    return ret



"""generators.

Networkx graph generators
"""
import itertools
import random
import uuid
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import networkx as nx
import torch

from caldera.utils import functional as Fn
from caldera.utils._tools import _resolve_range
from caldera.utils.nx import nx_is_directed


def nx_random_features(g: nx.DiGraph, n_feat: int, e_feat: int, g_feat: int):
    for _, ndata in g.nodes(data=True):
        ndata["features"] = torch.randn(n_feat)
    for _, _, edata in g.edges(data=True):
        edata["features"] = torch.randn(e_feat)
    g.data = {"features": torch.randn(g_feat)}
    return g


def rand_n_nodes_n_edges(
    n_nodes: Union[int, Tuple[int, int]],
    n_edges: Optional[Union[int, Tuple[int, int]]] = None,
    density: Optional[Union[float, Tuple[float, float]]] = None,
):
    n = int(n_nodes)
    if n_edges is None and density is None:
        raise ValueError("Either density or n_edges must be provided.")
    elif n_edges is None:
        d = float(density)
        e = int((d * n * (n - 1)) / 2)
        e = max(1, e)
    else:
        e = int(n_edges)
    return n, e


def random_graph(
    n_nodes: Union[int, Tuple[int, int]],
    n_edges: Optional[Union[int, Tuple[int, int]]] = None,
    density: Optional[Union[float, Tuple[float, float]]] = None,
    generator: Callable[[int, int], nx.Graph] = nx.generators.dense_gnm_random_graph,
    *args,
    **kwargs
):
    n, e = rand_n_nodes_n_edges(n_nodes, n_edges, density)
    return generator(n, e, *args, **kwargs)


def random_node(g, n=None):
    nodes = [random.choice(list(g.nodes)) for _ in range(n or 1)]
    if n is None:
        nodes = nodes[0]
    return nodes


def random_edge(g, n=None):
    edges = [random.choice(list(g.edges)) for _ in range(n or 1)]
    if n is None:
        edges = edges[0]
    return edges


def _possible_edges(n1: Set, n2: Set, directed: bool, self_loops: bool = False):
    """Compute the number of possible edges between two sets."""
    a = n1.intersection(n2)
    e = (len(n1) - len(a)) * (len(n2) - len(a))
    if directed:
        e *= 2
    if self_loops:
        e += len(n1) + len(n2) - len(a)
    return e


def connect_node_sets(
    g,
    s1,
    s2,
    density: Union[float, Tuple[float, float]],
    edge_data: Optional[Dict] = None,
):
    """Connect two node sets at the specified density."""
    if not isinstance(g, nx.Graph):
        raise TypeError(
            "`g` must be a `nx.Graph` subclass, not a `{}`".format(g.__class__)
        )
    if not isinstance(s1, set):
        raise TypeError("node set `s1` must be a `set`, not a `{}`".format(g.__class__))
    if not isinstance(s2, set):
        raise TypeError("node set `s2` must be a `set`, not a `{}`".format(g.__class__))
    m1 = s1.difference(set(g))
    if m1:
        raise ValueError("Nodes missing from graph. " + str(m1))

    m2 = s2.difference(set(g))
    if m2:
        raise ValueError("Nodes missing from graph. " + str(m2))

    edge_data = edge_data or dict()
    existing_edges = set()
    n_possible_edges = _possible_edges(
        s1, s2, directed=nx_is_directed(g), self_loops=False
    )
    for n1, n2 in g.edges():
        if (n1 in s1 and n2 in s2) or (n2 in s1 and n1 in s2):
            existing_edges.add((n1, n2))
            if not nx_is_directed:
                existing_edges.add((n2, n1))
    n_new_edges = max(0, int((n_possible_edges * density) - len(existing_edges)))

    add_new_edges = Fn.compose(
        Fn.map_each(list),
        Fn.apply_each(Fn.shuffle_each()),  # shuffle nodes
        lambda arr: itertools.product(*list(arr)),  # all possible edges
        Fn.filter_each(lambda x: x not in existing_edges),  # only new edges
        Fn.apply_each(lambda x: g.add_edge(x[0], x[1], **edge_data)),  # add edge
        Fn.iter_count(n_new_edges),  # limit number of new edges,
        list,
    )

    add_new_edges((s1, s2))

    return g


def compose_and_connect(
    g: nx.Graph,
    h: nx.Graph,
    density: Union[float, Tuple[float, float]],
    edge_data: Optional[Dict] = None,
) -> nx.Graph:
    """Compose two graphs and connect them at the given density.

    With density being `(existing edges between G, H) / (possible edges
    between G, H)`
    """
    i = nx.compose(g, h)
    density = _resolve_range(density)
    connect_node_sets(i, set(g), set(h), density, edge_data)
    return i


def chain_graph(sequence, graph_class, edge_data: Optional[Dict] = None):
    g = graph_class()
    edge_data = edge_data or dict()
    for n1, n2 in nx.utils.pairwise(sequence):
        g.add_edge(n1, n2, **edge_data)
    return g


uuid_sequence = Fn.compose(
    lambda x: range(x), Fn.map_each(lambda _: str(uuid.uuid4())[-5:])
)


def unique_chain_graph(n: int, graph_class, edge_data: Optional[Dict] = None):
    return chain_graph(uuid_sequence(n), graph_class, edge_data)



from abc import ABC
from functools import wraps
from typing import List
from typing import Tuple
from typing import Union

import torch
import torch_scatter
from torch import nn

from caldera.defaults import CalderaDefaults as D


@wraps(torch_scatter.scatter_max)
def scatter_max(*args, **kwargs):
    return torch_scatter.scatter_max(*args, **kwargs)[0]


@wraps(torch_scatter.scatter_min)
def scatter_min(*args, **kwargs):
    return torch_scatter.scatter_min(*args, **kwargs)[0]


class AggregatorBase(nn.Module, ABC):
    """Aggregation layer."""

    valid_aggregators = {
        "mean": torch_scatter.scatter_mean,
        "max": scatter_max,
        "min": scatter_min,
        "add": torch_scatter.scatter_add,
    }


# TODO: make aggregation selection trainable
class Aggregator(AggregatorBase):
    """Aggregation layer."""

    def __init__(self, aggregator: str, dim: int = None, dim_size: int = None):
        super().__init__()

        if aggregator not in self.valid_aggregators:
            raise ValueError(
                "Aggregator '{}' not not one of the valid aggregators {}".format(
                    aggregator, self.valid_aggregators
                )
            )
        self.aggregator = aggregator
        self.kwargs = dict(dim=dim, dim_size=dim_size)

    def forward(self, x, indices, **kwargs):
        func_kwargs = dict(self.kwargs)
        func_kwargs.update(kwargs)
        func = self.valid_aggregators[self.aggregator]
        result = func(x, indices, **func_kwargs)
        return result

    @staticmethod
    @wraps(torch_scatter.scatter_max)
    def scatter_max(*args, **kwargs):
        return torch_scatter.scatter_max(*args, **kwargs)[0]

    @staticmethod
    @wraps(torch_scatter.scatter_min)
    def scatter_min(*args, **kwargs):
        return torch_scatter.scatter_min(*args, **kwargs)[0]

    def __repr__(self):
        return "{}(func='{}')".format(self.__class__.__name__, self.aggregator)


# TODO: This is where the network decides the type of causal relationship.
#       Exposing this could help with network 'explainability'
#       You could apply a loss function so the network is encouraged
#       to pick a single explanation for the data.


class MultiAggregator(AggregatorBase):
    def __init__(
        self,
        input_size: int,
        aggregators: Union[Tuple[str, ...], List[str]],
        dim: int = None,
        dim_size: int = None,
        activation=D.activation,
    ):
        """A differentiable and trainable way to select the aggregation
        function.

        :param input_size:
        :param aggregators:
        :param dim:
        :param dim_size:
        """
        super().__init__()
        for aggregator in aggregators:
            if aggregator not in self.valid_aggregators:
                raise ValueError(
                    "Aggregator '{}' not not one of the valid aggregators {}".format(
                        aggregator, self.valid_aggregators
                    )
                )
        self.layers = torch.nn.Sequential(
            torch.nn.Linear(input_size, len(aggregators)), activation()
        )
        self.aggregators = torch.nn.ModuleDict(
            {agg: Aggregator(agg) for agg in aggregators}
        )

        self.kwargs = dict(dim=dim, dim_size=dim_size)

    def forward(self, x, indices, **kwargs):
        func_kwargs = dict(self.kwargs)
        func_kwargs.update(kwargs)

        # stack each aggregation function
        stacked = torch.stack(
            [agg(x, indices, **func_kwargs) for agg in self.aggregators.values()]
        )

        # get the weights
        weights = self.layers(x)

        # match shape of aggregated matrix
        scatter_weights = self.valid_aggregators["add"](weights, indices, **func_kwargs)

        # weight each 'function' by the learned weights
        return torch.sum(
            torch.mul(stacked, scatter_weights.expand(1, -1, -1).T), axis=0
        )



from typing import Callable
from typing import TypeVar

import networkx as nx

from ._base import NetworkxTransformBase

T = TypeVar("T")


class NetworkxApply(NetworkxTransformBase):
    def __init__(self, func: Callable[[nx.Graph], T]):
        self.func = func

    def transform(self, g: nx.Graph) -> T:
        return self.func(g)



from typing import Union

import networkx as nx

Graph = Union[
    nx.DiGraph,
    nx.Graph,
    nx.MultiGraph,
    nx.MultiDiGraph,
    nx.OrderedDiGraph,
    nx.OrderedGraph,
]
DirectedGraph = Union[
    nx.DiGraph, nx.MultiDiGraph, nx.OrderedDiGraph, nx.OrderedMultiDiGraph
]
UndirectedGraph = Union[nx.Graph, nx.MultiGraph, nx.OrderedGraph, nx.OrderedMultiGraph]



import torch
from torch import nn

from caldera.blocks.aggregator import Aggregator
from caldera.blocks.block import Block
from caldera.data import GraphBatch


class NodeBlock(Block):
    def __init__(self, mlp: nn.Module):
        super().__init__({"mlp": mlp}, independent=True)

    def forward(self, node_attr):
        return self.block_dict["mlp"](node_attr)

    def forward_from_data(self, data: GraphBatch):
        return self(data.x)


class AggregatingNodeBlock(NodeBlock):
    def __init__(self, mlp: nn.Module, edge_aggregator: Aggregator):
        super().__init__(mlp)
        self.block_dict["edge_aggregator"] = edge_aggregator
        self._independent = False

    # TODO: source_to_dest or dest_to_source (isn't this just reversing the graph?)
    def forward(
        self,
        *,
        node_attr,
        edge_attr,
        edges,
        global_attr: torch.Tensor = None,
        node_idx: torch.Tensor = None,
    ):
        aggregated = (
            self.block_dict["edge_aggregator"](
                edge_attr, edges[1], dim=0, dim_size=node_attr.size(0)
            ),
        )
        if global_attr is not None:
            if node_idx is None:
                raise RuntimeError(
                    "Must provide `node_index` if providing `global_attr`"
                )
            aggregated += (global_attr[node_idx],)
        out = torch.cat([node_attr, *aggregated], dim=1)
        return self.block_dict["mlp"](out)

    def forward_from_data(self, data: GraphBatch):
        return self(data.x, data.e, data.edges)



"""_fully_connected.py.

Make graph fully connected
"""
from typing import overload

from ._base import TransformBase
from caldera.data import GraphBatch
from caldera.data import GraphData
from caldera.data.utils import add_edges


class FullyConnected(TransformBase):
    def __init__(self, fill_value: float = 0.0):
        super().__init__()
        self.fill_value = fill_value

    # def __call__(self, data: GraphData) -> GraphData:
    #     if issubclass(data.__class__, GraphBatch):
    #         return data.__class__.from_data_list([self(d) for d in data.to_data_list()])
    #     with torch.no_grad():
    #         all_graph_edges = _create_all_edges(0, data.num_nodes)
    #         missing_edges = edges_difference(all_graph_edges, data.edges)
    #         edges = torch.cat([data.edges, missing_edges], axis=1)
    #         edge_attr = torch.full((edges.shape[1], data.e.shape[1]), fill_value=self.fill_value)
    #     return GraphData(
    #         node_attr=data.x.detach().clone(),
    #         edge_attr=edge_attr,
    #         global_attr=data.g.detach().clone(),
    #         edges=edges
    #     )

    @overload
    def transform_each(self, data: GraphData) -> GraphData:
        ...

    def transform(self, data: GraphBatch) -> GraphBatch:
        return add_edges(data, self.fill_value, kind="complete")



from os.path import isdir

try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    SummaryWriter = None

if SummaryWriter is None:

    def new_writer(*args, **kwargs):
        raise ImportError("`tensorboard` not installed")


else:

    def new_writer(directory: str, suffix=""):
        i = 0

        def name(index):
            return directory + "%04d" % index + suffix

        while isdir(name(i)):
            i += 1
        dirname = name(i)
        print("New writer at '{}'".format(dirname))
        return SummaryWriter(dirname)



from dataclasses import dataclass
from dataclasses import field
from typing import Optional
from typing import Tuple

import torch
from torch.nn import LeakyReLU

from .tools import ConfigObj


@dataclass
class LayerConfig(ConfigObj):
    """Configuration for an edge, node, or global graph layer."""

    size: int = 1
    depth: int = 1
    layer_norm: bool = "${network.layer_norm}"
    activation: Optional[str] = "${network.activation}"
    dropout: float = "${network.dropout}"

    @property
    def layers(self):
        return [self.size] * self.depth


@dataclass
class NetComponentConfig(ConfigObj):

    node: LayerConfig = field(default_factory=LayerConfig)
    edge: LayerConfig = field(default_factory=LayerConfig)
    glob: LayerConfig = field(default_factory=LayerConfig)


@dataclass
class NetConfig(ConfigObj):
    """Configuration for caldera Graph Network."""

    encode: NetComponentConfig = field(default_factory=NetComponentConfig)
    core: NetComponentConfig = field(default_factory=NetComponentConfig)
    out: NetComponentConfig = field(default_factory=NetComponentConfig)

    # defaults
    dropout: Optional[float] = None
    activation: Optional[str] = LeakyReLU.__name__
    layer_norm: bool = True

    aggregator_activation: str = LeakyReLU.__name__

    # architecture
    pass_global_to_edge: bool = True
    pass_global_to_node: bool = True

    # aggregators
    edge_block_to_node_aggregators: Tuple[str, ...] = tuple(["add"])
    global_block_to_node_aggregators: Tuple[str, ...] = tuple(["add"])
    global_block_to_edge_aggregators: Tuple[str, ...] = tuple(["add"])

    @staticmethod
    def get_activation(activation: str):
        if activation is None:
            return torch.nn.Identity
        return getattr(torch.nn, activation)



"""_random_edge_mask.py.

Apply random edge mask to the graph.
"""
from typing import overload

import torch
from torch import distributions

from ._base import TransformBase
from caldera.data import GraphBatch
from caldera.data import GraphData


class RandomEdgeMask(TransformBase):
    def __init__(self, dropout: float):
        if dropout < 0.0 or dropout > 1.0:
            raise ValueError("Edge dropout rate must be between [0., 1.]")
        self.dist = distributions.Bernoulli(1.0 - dropout)

    @overload
    def transform_each(self, data: GraphData) -> GraphData:
        ...

    def transform(self, data: GraphBatch) -> GraphBatch:
        mask = self.dist.sample((data.num_edges,)).to(torch.bool)
        return data.apply_edge_mask(mask)
